'''
==========================================================================
test_sinks.py
==========================================================================
A collection of test sinks.

Author : Yanghui Ou
  Date : Feb 3, 2020
'''
from pymtl3 import *
from pymtl3.stdlib.ifcs import RecvRTL2SendCL, RecvIfcRTL

from ..utils import get_nbits
from ..packets import MflitPacket as Packet

#-------------------------------------------------------------------------
# MflitPacketSinkCL
#-------------------------------------------------------------------------

class MflitPacketSinkCL( Component ):

  def construct( s, Format, pkts, initial_delay=0, flit_interval_delay=0,
                 packet_interval_delay=0, cmp_fn=lambda a, b : a.flits == b.flits ):

    s.PhitType  = mk_bits( get_nbits( Format ) )
    s.recv.Type = s.PhitType
    s.Format    = Format

    s.cur_pkt  = Packet( Format )
    s.ref_pkts = list( pkts )
    s.count    = initial_delay

    s.cmp_fn    = cmp_fn
    s.error_msg = ''

    s.all_received = False
    s.done_flag    = False
    s.recv_called  = False

    @s.update
    def up_sink_count():
      # Raise exception at the start of next cycle so that the errored
      # line trace gets printed out
      if s.error_msg:
        raise Exception( s.error_msg )

      # Tick one more cycle after message is received so that the
      # exception gets thrown
      if s.all_received:
        s.done_flag = True

      if len( s.ref_pkts ) == 0:
        s.all_received = True

      # If recv was called in previous cycle
      if s.recv_called:
        if s.cur_pkt.empty():
          s.count = packet_interval_delay
        else:
          s.count = flit_interval_delay

      elif s.count != 0:
        s.count -= 1

      else:
        s.count = 0

      s.recv_called = False

    # Constraints
    s.add_constraints(
      U( up_sink_count ) < M( s.recv     ),
      U( up_sink_count ) < M( s.recv.rdy ),
    )

  @non_blocking( lambda s: s.count==0 )
  def recv( s, flit ):
    assert s.count == 0
    s.recv_called = True

    # Received more packets
    if len( s.ref_pkts ) == 0:
      s.error_msg = (
        f'Test sink {s} received more flits than expected!\n'
        f'Received : {flit}'
      )

    # Assemble packet
    # FIXME: use clone() to avoid mutable issue once pymtl3 is updated
    s.cur_pkt.add( s.PhitType(flit) )

    # If a packet is assembled, check if it exists in the reference
    if s.cur_pkt.full():

      # Check
      flag = False
      for i, pkt in enumerate( s.ref_pkts ):
        if s.cmp_fn( s.cur_pkt, pkt ):
          flag = True
          s.ref_pkts.pop( i )
          break

      if not flag:
        s.error_msg = (
          f'Test sink {s} received unexpected packet!\n'
          f'Received : {s.cur_pkt.flits}\n'
          f'Expecting: {[ p.flits for p in s.ref_pkts]}'
        )

      # Reset current packet to empty
      s.cur_pkt = Packet( s.Format )

  def done( s ):
    return s.done_flag

  def line_trace( s ):
    return f'{s.recv}({s.count})'

#-------------------------------------------------------------------------
# MflitPacketSinkRTL
#-------------------------------------------------------------------------

class MflitPacketSinkRTL( Component ):

  def construct( s, Format, pkts, initial_delay=0, flit_interval_delay=0,
                 packet_interval_delay=0, cmp_fn=lambda a, b : a.flits == b.flits ):

    s.PhitType = mk_bits( get_nbits( Format ) )

    s.recv     = RecvIfcRTL( s.PhitType )
    s.sink_cl  = MflitPacketSinkCL( Format, pkts, initial_delay, flit_interval_delay,
                                         packet_interval_delay , cmp_fn )
    s.adapter = RecvRTL2SendCL( s.PhitType )

    connect( s.recv,         s.adapter.recv )
    connect( s.adapter.send, s.sink_cl.recv )

  def done( s ):
    return s.sink_cl.done()

  def line_trace( s ):
    return f'{s.recv}'



"""
=========================================================================
RingNetworkRTL_test.py
=========================================================================
Test for RingNetworkRTL

Author : Yanghui Ou, Cheng Tan
  Date : June 28, 2019
"""
from ocnlib.ifcs.packets import mk_ring_pkt
from ocnlib.ifcs.positions import mk_ring_pos
from ocnlib.utils import run_sim
from ocnlib.test.net_sinks import TestNetSinkRTL
from pymtl3 import *
from pymtl3.stdlib.test.test_srcs import TestSrcRTL
from ringnet.RingNetworkRTL import RingNetworkRTL

from ..RingNetworkFL import ringnet_fl

#-------------------------------------------------------------------------
# TestHarness
#-------------------------------------------------------------------------

class TestHarness( Component ):

  def construct( s, MsgType, num_routers, src_msgs, sink_msgs ):

    s.num_routers = num_routers
    RingPos = mk_ring_pos( num_routers )
    match_func = lambda a, b : a.payload == b.payload

    s.srcs  = [ TestSrcRTL( MsgType, src_msgs[i] )
              for i in range( num_routers ) ]
    s.dut   = RingNetworkRTL( MsgType, RingPos, num_routers, 0)
    s.sinks = [ TestNetSinkRTL( MsgType, sink_msgs[i],
              match_func = match_func )
              for i in range( num_routers ) ]

    # Connections
    for i in range ( s.dut.num_routers ):
      s.srcs[i].send //= s.dut.recv[i]
      s.dut.send[i]  //= s.sinks[i].recv

  def done( s ):
    srcs_done = True
    sinks_done = True
    for i in range( s.num_routers ):
      srcs_done = srcs_done and s.srcs[i].done()
      sinks_done = sinks_done and s.sinks[i].done()
    return srcs_done and sinks_done

  def line_trace( s ):
    return s.dut.line_trace()

#-------------------------------------------------------------------------
# mk_src_pkts
#-------------------------------------------------------------------------

def mk_src_pkts( nterminals, lst ):
  src_pkts = [ [] for _ in range( nterminals ) ]
  src = 0
  for pkt in lst:
    if hasattr(pkt, 'fl_type'):
      if pkt.fl_type == 0:
        src = pkt.src
    else:
      src = pkt.src
    src_pkts[ src ].append( pkt )
  return src_pkts

#=========================================================================
# Test cases
#=========================================================================

class RingNetwork_Tests:

  @classmethod
  def setup_class( cls ):
    cls.DutType = RingNetworkRTL

  def test_simple( s ):
    nterminals = 4
    Pkt = mk_ring_pkt( nterminals )
    src_pkts = mk_src_pkts( nterminals, [
      #    src  dst opq vc payload
      Pkt( 3,   0,  0,  0, 0xfaceb00c ),
    ])
    dst_pkts = ringnet_fl( src_pkts )
    th = TestHarness( Pkt, nterminals, src_pkts, dst_pkts )
    run_sim( th )

  def test_cycle( s ):
    nterminals = 4
    Pkt = mk_ring_pkt( nterminals )
    src_pkts = mk_src_pkts( nterminals, [
      #    src  dst opq vc payload
      Pkt( 0,   1,  0,  0, 0xfaceb00c ),
      Pkt( 1,   2,  1,  0, 0xdeadbeef ),
      Pkt( 2,   3,  2,  0, 0xbaadface ),
      Pkt( 3,   0,  0,  0, 0xfaceb00c ),
    ])
    dst_pkts = ringnet_fl( src_pkts )
    th = TestHarness( Pkt, nterminals, src_pkts, dst_pkts )
    run_sim( th )

  def test_anti_cycle( s ):
    nterminals = 4
    Pkt = mk_ring_pkt( nterminals )
    src_pkts = mk_src_pkts( nterminals, [
      #    src  dst opq vc payload
      Pkt( 0,   3,  0,  0, 0xfaceb00c ),
      Pkt( 1,   0,  1,  0, 0xdeadbeef ),
      Pkt( 2,   1,  2,  0, 0xbaadface ),
      Pkt( 3,   2,  0,  0, 0xfaceb00c ),
    ])
    dst_pkts = ringnet_fl( src_pkts )
    th = TestHarness( Pkt, nterminals, src_pkts, dst_pkts )
    run_sim( th )



"""
==========================================================================
 SwitchUnitCL_test.py
==========================================================================

Author: Yanghui Ou
  Date: July 2, 2019
"""
import pytest

from ocnlib.utils import run_sim
from ocnlib.test.net_sinks import TestNetSinkCL as TestSinkCL
from pymtl3 import *
from pymtl3.stdlib.cl.queues import BypassQueueCL
from pymtl3.stdlib.test.test_srcs import TestSrcCL
from router.SwitchUnitCL import SwitchUnitCL

#-------------------------------------------------------------------------
# TestHarness
#-------------------------------------------------------------------------

class TestHarness( Component ):

  def construct( s, MsgType, src_msgs, sink_msgs, num_inports=5 ):

    match_func = lambda a,b : a==b
    s.src   = [ TestSrcCL( MsgType, src_msgs[i] ) for i in range(num_inports) ]
    s.src_q = [ BypassQueueCL( num_entries=1 ) for _ in range(num_inports) ]
    s.dut   = SwitchUnitCL( MsgType, num_inports )
    s.sink  = TestSinkCL( MsgType, sink_msgs, match_func=match_func )

    # Connections
    for i in range( 5 ):
      s.src[i].send  //= s.src_q[i].enq
      s.src_q[i].deq //= s.dut.get[i]

    @s.update
    def up_give_recv():
      if s.dut.give.rdy() and s.sink.recv.rdy():
        s.sink.recv( s.dut.give() )

  def done( s ):
    srcs_done = True
    for i in range( 5 ):
      if not s.src[i].done():
        srcs_done = False
    return srcs_done and s.sink.done()

  def line_trace( s ):
    return "{}".format( s.dut.line_trace() )

#-------------------------------------------------------------------------
# Test cases
#-------------------------------------------------------------------------

class SwitchUnitCL_Tests:

  @classmethod
  def setup_class( cls ):
    cls.TestHarness = TestHarness

  def test_su5_simple( s ):
    src_msgs    = [ [] for _ in range(5) ]
    src_msgs[0] = [ b16(4), b16(5)  ]
    src_msgs[1] = [ b16(2), b16(8)  ]
    src_msgs[2] = [ b16(4), b16(13) ]
    src_msgs[3] = [ b16(6), b16(18) ]
    src_msgs[4] = [ b16(8), b16(23) ]
    sink_msg = [ msg for i in range(5) for msg in src_msgs[i]  ]
    th = s.TestHarness( Bits16, src_msgs, sink_msg, num_inports=5 )
    run_sim( th )



#=========================================================================
# DORYTorusRouteUnitCL.py
#=========================================================================
# A DOR route unit with get/give interface for Torus topology in CL.
#
# Author : Cheng Tan
#   Date : May 20, 2019

from directions import *
from pymtl3 import *


class DORYTorusRouteUnitCL( Component ):

  def construct( s,
                 PacketType,
                 PositionType,
                 num_outports,
                 cols=2,
                 rows=2 ):

    # Constants
    s.num_outports = num_outports
    s.cols = cols
    s.rows = rows

    # Interface

    s.get  = CallerIfcCL()
    s.give = [ CalleeIfcCL() for _ in range ( s.num_outports ) ]
    s.pos  = InPort( PositionType )

    # Componets

    s.rdy_lst = [ False for _ in range( s.num_outports ) ]
    s.msg     = None

    # Routing logic

    @s.update
    def up_ru_routing():

      if s.msg is None and s.get.rdy():
        s.msg = s.get()
      if s.msg is not None:
        if s.pos.pos_x == s.msg.dst_x and s.pos.pos_y == s.msg.dst_y:
          s.rdy_lst[SELF] = True
        elif s.msg.dst_y < s.pos.pos_y:
          if s.pos.pos_y - s.msg.dst_y <= s.rows - s.pos.pos_y + s.msg.dst_y:
            s.rdy_lst[SOUTH] = True
          else:
            s.rdy_lst[NORTH] = True
        elif s.msg.dst_y > s.pos.pos_y:
          if s.msg.dst_y - s.pos.pos_y <= s.rows - s.msg.dst_y + s.pos.pos_y:
            s.rdy_lst[NORTH] = True
          else:
            s.rdy_lst[SOUTH] = True
        elif s.msg.dst_x < s.pos.pos_x:
          if s.pos.pos_x - s.msg.dst_x <= s.cols - s.pos.pos_x + s.msg.dst_x:
            s.rdy_lst[WEST] = True
          else:
            s.rdy_lst[EAST] = True
        else:
          if s.msg.dst_x - s.pos.pos_x <= s.rows - s.msg.dst_x + s.pos.pos_x:
            s.rdy_lst[EAST] = True
          else:
            s.rdy_lst[WEST] = True
      else:
        s.rdy_lst = [ False for _ in range( s.num_outports ) ]

    # Assign method and ready

    for i in range( s.num_outports ):
      def gen_give_rdy( s, port_id ):
        def give_rdy():
          if s.msg is not None:
            return s.rdy_lst[port_id]
          else:
            return False
        return give_rdy

      s.give[i].rdy.method = gen_give_rdy( s, i )
      s.give[i].method.method = s.give_method

    for i in range( s.num_outports ):
      s.add_constraints(
        M( s.get ) < U( up_ru_routing ) < M( s.give[i] ),
      )

  def give_method( s ):
    assert s.msg is not None
    ret = s.msg
    s.msg = None
    return ret

   # TODO: CL line trace

  def line_trace( s ):
    out_str = "|".join([ str(s.give[i]) for i in range( s.num_outports ) ])
    return f"{s.get}(){out_str}"



"""
==========================================================================
TorusRouterFL.py
==========================================================================
FL route unit that implements dimension order routing.

Author : Yanghui Ou
  Date : June 30, 2019
"""
from pymtl3 import *

from .directions import *
from .RouteUnitDorFL import RouteUnitDorFL


class TorusRouterFL:

  def __init__( s, pos_x, pos_y, ncols, nrows, dimension='y' ):
    s.pos_x = pos_x
    s.pos_y = pos_y
    s.ncols = ncols
    s.nrows = nrows
    s.dimension  = dimension
    s.route_unit = RouteUnitDorFL( pos_x, pos_y, ncols, nrows, dimension='y' )

  #-----------------------------------------------------------------------
  # arrange_src_pkts
  #-----------------------------------------------------------------------
  # A helper function that puts each packet in [lst] into corresponding
  # source.

  def arrange_src_pkts( s, lst ):
    src_pkts = [ [] for _ in range(5) ]

    for pkt in lst:
      if pkt.src_x == s.pos_x and pkt.src_y == s.pos_y:
        in_dir = SELF

      elif s.dimension=='y':

        # Same x - either comes from north or south
        if pkt.src_x == s.pos_x:
          north_dist = pkt.dst_y - pkt.src_y if pkt.dst_y > pkt.src_y else pkt.dst_y + s.nrows - pkt.src_y
          south_dist = pkt.src_y - pkt.dst_y if pkt.dst_y < pkt.src_y else pkt.src_y + s.nrows - pkt.dst_y
          in_dir = SOUTH if north_dist < south_dist else NORTH

        # Different x - either comes from west or east
        else:
          east_dist  = pkt.dst_x - pkt.src_x if pkt.dst_x > pkt.src_x else pkt.dst_x + s.ncols - pkt.src_x
          west_dist  = pkt.src_x - pkt.dst_x if pkt.dst_x < pkt.src_x else pkt.src_x + s.ncols - pkt.dst_x
          in_dir = EAST if west_dist < east_dist else WEST

      else: # s.dimension=='x'

        # Same y - either comes from west or east
        if pkt.src_x == s.pos_x:
          east_dist  = pkt.dst_x - pkt.src_x if pkt.dst_x > pkt.src_x else pkt.dst_x + s.ncols - pkt.src_x
          west_dist  = pkt.src_x - pkt.dst_x if pkt.dst_x < pkt.src_x else pkt.src_x + s.ncols - pkt.dst_x
          in_dir = EAST if west_dist < east_dist else WEST

        # Different y - either comes from north or south
        else:
          north_dist = pkt.dst_y - pkt.src_y if pkt.dst_y > pkt.src_y else pkt.dst_y + s.nrows - pkt.src_y
          south_dist = pkt.src_y - pkt.dst_y if pkt.dst_y < pkt.src_y else pkt.src_y + s.nrows - pkt.dst_y
          in_dir = SOUTH if north_dist < south_dist else NORTH

      src_pkts[ in_dir ].append( pkt )

    return src_pkts

  #-----------------------------------------------------------------------
  # route
  #-----------------------------------------------------------------------
  # Use FL route unit to route each packet in [src_pkts] to corresponding
  # destination.

  def route( s, src_pkts ):
    assert len( src_pkts ) == 5

    dst_pkts = [ [] for _ in range(5) ]

    for pkts in src_pkts:
      tmp = s.route_unit.route( pkts )
      for i in range(5):
        dst_pkts[i].extend( tmp[i] )

    return dst_pkts



'''
==========================================================================
GrantHoldArbiter_test.py
==========================================================================

Author : Yanghui Ou
  Date : Jan 22, 2020
'''
from pymtl3 import *
from .GrantHoldArbiter import GrantHoldArbiter


def test_simple():
  arb = GrantHoldArbiter( nreqs=4 )
  arb.elaborate()
  arb.apply( SimulationPass() )
  arb.sim_reset()
  arb.tick()

  print()
  arb.reqs = b4(0b0011)
  arb.hold = b1(0)
  arb.eval_combinational()
  print( arb.line_trace() )
  arb.tick()
  
  arb.eval_combinational()
  print( arb.line_trace() )
  g0 = b4(arb.grants)
  arb.tick()

  arb.hold = b1(1)
  arb.eval_combinational()
  print( arb.line_trace() )
  assert arb.grants == g0
  arb.tick()




'''
==========================================================================
SerializerRTL_test.py
==========================================================================
Unit tests for SerializerRTL.

Author : Yanghui Ou
  Date : Feb 26, 2020
'''
import hypothesis
from hypothesis import strategies as st
from pymtl3.datatypes.strategies import bits 
from pymtl3 import *
from pymtl3.stdlib.test.test_srcs import TestSrcRTL as TestSource
from pymtl3.stdlib.test.test_sinks import TestSinkRTL as TestSink
from ocnlib.utils import run_sim
# from pymtl3.stdlib.test import run_sim

from .SerializerRTL import SerializerRTL

#-------------------------------------------------------------------------
# mk_msgs
#-------------------------------------------------------------------------

def mk_msgs( out_nbits, max_nblocks, msgs ):
  InType    = mk_bits( out_nbits*( max_nblocks ) )
  OutType   = mk_bits( out_nbits )
  LenType   = mk_bits( clog2( max_nblocks+1 ) )
  src_msgs  = [ InType(x) for x in msgs[::2] ]
  len_msgs  = [ LenType(x) for x in msgs[1::2] ]

  sink_msgs = []
  for data, length in zip( src_msgs, len_msgs ):
    assert length.uint() > 0
    for i in range( length.uint() ):
      sink_msgs.append( data[i*out_nbits:(i+1)*out_nbits] )

  return src_msgs, len_msgs, sink_msgs

#-------------------------------------------------------------------------
# TestHarness
#-------------------------------------------------------------------------

class TestHarness( Component ):
  def construct( s, out_nbits, max_nblocks, msgs ):
    InType  = mk_bits( out_nbits*( max_nblocks ) )
    OutType = mk_bits( out_nbits )
    LenType =  mk_bits( clog2( max_nblocks+1 ) )
    src_msgs, len_msgs, sink_msgs = mk_msgs( out_nbits, max_nblocks, msgs )

    s.src     = TestSource( InType, src_msgs )
    s.src_len = TestSource( LenType, len_msgs )
    s.dut     = SerializerRTL( out_nbits, max_nblocks )
    s.sink    = TestSink( OutType, sink_msgs )

    connect( s.src.send,         s.dut.recv     )
    connect( s.src_len.send.rdy, s.dut.recv.rdy )
    connect( s.src_len.send.msg, s.dut.len      )
    connect( s.dut.send,         s.sink.recv    )

  def done( s ):
    return s.src.done() and s.sink.done()

  def line_trace( s ):
    return s.dut.line_trace()

#-------------------------------------------------------------------------
# test case: sanity check
#-------------------------------------------------------------------------

def test_sanity_check():
  dut = SerializerRTL( out_nbits=16, max_nblocks=4 )
  dut.elaborate()
  dut.apply( SimulationPass() )
  dut.sim_reset()
  dut.tick()
  dut.tick()

  th = TestHarness( 16, 8, [] )
  th.elaborate()
  th.apply( SimulationPass() )
  th.tick()
  th.tick()

#-------------------------------------------------------------------------
# test case: basic
#-------------------------------------------------------------------------

def test_basic( test_verilog ):
  msgs = [
    0xfaceb00c,         2,
    0xdeadbeef,         1,
    0xcafebabefa11deaf, 4,
    0xace3ace2ace1ace0, 3,
  ]
  th = TestHarness( 16, 4, msgs )
  translation = 'verilog' if test_verilog else ''
  run_sim( th, max_cycles=40, translation=translation )

#-------------------------------------------------------------------------
# test case: backpressure
#-------------------------------------------------------------------------

def test_backpressure( test_verilog ):
  msgs = [
    0xfaceb00c,         2,
    0xdeadbeef,         1,
    0xcafebabefa11deaf, 4,
    0xace3ace2ace1ace0, 3,
  ]
  th = TestHarness( 16, 4, msgs )
  th.set_param( 'top.sink.construct', initial_delay=10, interval_delay=2 )
  translation = 'verilog' if test_verilog else ''
  run_sim( th, max_cycles=40, translation=translation )

#-------------------------------------------------------------------------
# test case: src delay
#-------------------------------------------------------------------------

def test_src_delay( test_verilog ):
  msgs = [
    0xdeadbeef_faceb00c,                   2,
    0xbad0bad0_deadbeef,                   1,
    0xcafebabe_fa11deaf_deadc0de_faceb00c, 4,
    0xace5ace4_ace3ace2_ace1ace0,          3,
  ]
  th = TestHarness( 32, 9, msgs )
  th.set_param( 'top.src*.construct', initial_delay=10, interval_delay=2 )
  translation = 'verilog' if test_verilog else ''
  run_sim( th, max_cycles=40, translation=translation )

#-------------------------------------------------------------------------
# test case: stream 
#-------------------------------------------------------------------------

def test_stream( test_verilog ):
  msgs = [
    0xdeadbeef, 1,
    0xbad0bad0, 1,
    0xcafebabe, 1,
    0xace5ace4, 1,
    0xfaceb00c_ace5ace4, 2,
    0x8badf00d_ace5ace4, 2,
    0x8badc0de_ace5ace4, 2,
    0xfeedbabe_ace5ace4, 2,
  ]
  th = TestHarness( 32, 4, msgs )
  translation = 'verilog' if test_verilog else ''
  run_sim( th, max_cycles=40, translation=translation )

#-------------------------------------------------------------------------
# test case: pyh2
#-------------------------------------------------------------------------

@hypothesis.settings( deadline=None, max_examples=100 )
@hypothesis.given(
  out_nbits   = st.integers(1, 64),
  max_nblocks = st.integers(2, 15),
  data        = st.data(),
)
def test_pyh2( out_nbits, max_nblocks, data, test_verilog ):
  len_msgs = data.draw( st.lists( st.integers(1, max_nblocks), min_size=1, max_size=100 ) )
  src_msgs = [ data.draw( bits(x*out_nbits) ) for x in len_msgs ]

  msgs = []
  for x, l in zip( src_msgs, len_msgs ):
    msgs.append( x )
    msgs.append( l )

  th = TestHarness( out_nbits, max_nblocks, msgs )
  translation = 'verilog' if test_verilog else ''
  run_sim( th, translation=translation )




#=========================================================================
# RingRouteUnitCL.py
#=========================================================================
# Cycle level implementation of the route unit for ring network.
# It uses greedy routing algorithm.
#
# Author : Yanghui Ou
#   Date : May 16, 2019

from directions import *
from pymtl3 import *
from pymtl3.stdlib.cl.queues import BypassQueueCL


class RingRouteUnitCL( Component ):

  def construct( s, PacketType, PositionType, num_routers=4 ):

    # Constants

    s.num_outports = 3 # left, right, self
    s.total_dist = num_routers-1

    # Interface

    s.get  = CallerIfcCL()
    s.give = [ CalleeIfcCL() for _ in range ( s.num_outports ) ]
    s.pos  = InPort( PositionType )

    # Components

    s.rdy_lst = [ False for _ in range( s.num_outports ) ]
    s.msg     = None

    @s.update
    def ru_up_route():
      if s.msg is None and s.get.rdy():
        s.msg = s.get()

      s.rdy_lst = [ False for _ in range( s.num_outports ) ]
      if s.msg is not None:
        if s.msg.dst == s.pos:
          s.rdy_lst[SELF] = True
        else:
          if s.msg.dst > s.pos:
            right_dist = s.msg.dst - s.pos
            left_dist  = s.total_dist - right_dist
          else:
            left_dist  = s.msg.dst - s.pos
            right_dist = s.total_dist - left_dist
          if left_dist < right_dist:
            s.rdy_lst[LEFT] = True
          else:
            s.rdy_lst[RIGHT] = True

    # Assign method and ready

    for i in range( s.num_outports ):
      def gen_give_rdy( s, port_id ):
        def give_rdy():
          if s.msg is not None:
            return s.rdy_lst[port_id]
          else:
            return False
        return give_rdy

      s.give[i].rdy.method = gen_give_rdy( s, i )
      s.give[i].method.method = s.give_method

    for i in range( s.num_outports ):
      s.add_constraints(
        M( s.get ) < U( ru_up_route ) < M( s.give[i] ),
      )

  def give_method( s ):
    assert s.msg is not None
    ret = s.msg
    s.msg = None
    return ret

  # TODO: CL line trace

  def line_trace( s ):
    return "{!s:12}".format( s.msg )



"""
==========================================================================
 Counter.py
==========================================================================
A simple counter implementation.

Author : Yanghui Ou
  Date : June 13, 2019
"""

from pymtl3 import *


class Counter( Component ):

  def construct( s, Type, reset_value=0 ):

    # Interface

    s.incr       = InPort ( Bits1 )
    s.decr       = InPort ( Bits1 )
    s.load       = InPort ( Bits1 )
    s.load_value = InPort ( Type  )
    s.count      = OutPort( Type  )

    # Logic

    @s.update_ff
    def up_count():

      if s.reset:
        s.count <<= Type( reset_value )

      elif s.load:
        s.count <<= s.load_value

      elif s.incr & ~s.decr:
        s.count <<= s.count + Type(1)

      elif ~s.incr & s.decr:
        s.count <<= s.count - Type(1)

  def line_trace( s ):
    return "{}".format( s.count )



"""
==========================================================================
utils.py
==========================================================================
Utility functions for pyocn script.

Author: Yanghui Ou
  Date: Sep 24, 2019

"""
from pymtl3 import *
from pymtl3.stdlib.cl.queues import BypassQueueCL


class CLNetWrapper( Component ):

  def construct( s, PktType, net, nports ):

    # Interface

    s.recv = [ CalleeIfcCL( PktType ) for _ in range( nports ) ]
    s.give = [ CalleeIfcCL( PktType ) for _ in range( nports ) ]

    # Component

    s.net   = net
    s.out_q = [ BypassQueueCL( num_entries=1 ) for _ in range( nports ) ]

    for i in range( nports ):
      connect( s.recv[i],      s.net.recv[i]  )
      connect( s.net.send[i],  s.out_q[i].enq )
      connect( s.out_q[i].deq, s.give[i]      )

  def line_trace( s ):
    return s.net.line_trace()



'''
==========================================================================
BoundaryUnit.py
==========================================================================
Error check unit that solves the dangling caller port issue.

Author : Yanghui Ou
  Date : Feb 17, 2020
'''
from pymtl3 import *

class BoundaryUnit( Component ):

  def construct( s, default_rdy=False ):
    s.default_rdy = default_rdy

  @non_blocking( lambda s: s.default_rdy )
  def recv( s, *args, **kwargs ):
    assert False, f'Boundary unit {s} is not supposed to receive any message!'



"""
==========================================================================
SwitchUnitCL.py
==========================================================================
Cycle level implementation of a round robin switch unit.

Author : Yanghui Ou
  Date : May 16, 2019
"""
from pymtl3 import *


class SwitchUnitCL( Component ):

  def construct( s, PacketType, num_inports=5 ):

    # Constants

    s.num_inports = num_inports

    # Interface

    s.get  = [ CallerIfcCL( Type=PacketType ) for _ in range( s.num_inports ) ]
    s.give.Type = PacketType

    # Components

    s.priority = [ i for i in range( s.num_inports ) ]
    s.msg = None

    for i in range( s.num_inports ):
      s.add_constraints( M( s.get[i] ) == M( s.give ) )

  def any_ready( s ):
    flag = False
    for i in range( s.num_inports ):
      if s.get[i].rdy is not None:
        flag = flag or s.get[i].rdy()
    return flag

  @non_blocking( lambda s: s.any_ready() )
  def give( s ):
    for i in s.priority:
      if s.get[i].rdy():
        s.priority.remove( i )
        s.priority.append( i )
        return s.get[i]()

  # CL line trace

  def line_trace( s ):
    in_str = "|".join([ str(s.get[i]) for i in range(s.num_inports) ])
    return f"{in_str}_({s.priority})_{s.give}"



"""
==========================================================================
 Counter_test.py
==========================================================================
Tests for simple counter.

Author : Yanghui Ou
  Date : June 13, 2019
"""

from pymtl3 import *

from .Counter import Counter


def test_simple():
  print()
  dut = Counter( Bits4, reset_value=9 )
  dut.apply( SimulationPass() )
  dut.incr = b1(0)
  dut.decr = b1(0)
  dut.load = b1(0)
  dut.load = b4(0)

  dut.sim_reset()
  print( dut.line_trace() )
  dut.incr = b1(1)
  dut.eval_combinational()
  dut.tick()
  print( dut.line_trace() )
  assert dut.count == b4(10)
  dut.tick()
  print( dut.line_trace() )
  assert dut.count == b4(11)
  dut.tick()
  print( dut.line_trace() )
  assert dut.count == b4(12)

  dut.decr = b1(1)
  dut.eval_combinational()
  dut.tick()
  print( dut.line_trace() )
  assert dut.count == b4(12)

  dut.incr = b1(0)
  dut.eval_combinational()
  dut.tick()
  print( dut.line_trace() )
  assert dut.count == b4(11)
  dut.tick()
  print( dut.line_trace() )
  assert dut.count == b4(10)
  dut.tick()
  print( dut.line_trace() )
  assert dut.count == b4(9)

def _test_load():
  print
  dut = Counter( Bits4, reset_value=9 )
  dut.apply( SimulationPass() )
  dut.incr = b1(0)
  dut.decr = b1(0)
  dut.load = b1(0)
  dut.load = b4(0)

  dut.sim_reset()
  print( dut.line_trace() )
  dut.incr = b1(1)
  dut.load = b1(1)
  dut.load_value = b4(3)
  dut.eval_combinational()
  dut.tick()
  print( dut.line_trace() )
  assert dut.count == b4(3)
  dut.load = b1(0)
  dut.eval_combinational()
  dut.tick()
  print( dut.line_trace() )
  assert dut.count == b4(4)
  dut.tick()
  print( dut.line_trace() )
  assert dut.count == b4(5)

  dut.decr = b1(1)
  dut.eval_combinational()
  dut.tick()
  print( dut.line_trace() )
  assert dut.count == b4(5)

  dut.incr = b1(0)
  dut.eval_combinational()
  dut.tick()
  print( dut.line_trace() )
  assert dut.count == b4(4)
  dut.tick()
  print( dut.line_trace() )
  assert dut.count == b4(3)
  dut.tick()
  print( dut.line_trace() )
  assert dut.count == b4(2)



#=========================================================================
# ChannelCL.py
#=========================================================================
# CL channel module for connecting routers to form network. This simple
# channel has latency insensitive send/recv interfaces.
#
# Author : Yanghui Ou
#   Date : May 19, 2019

from pymtl3 import *
from pymtl3.stdlib.cl.queues import NormalQueueCL


class ChannelCL( Component ):

  def construct( s, PacketType, QueueType = NormalQueueCL, latency = 0 ):

    # Interface

    s.recv = CalleeIfcCL()
    s.send = CallerIfcCL()
    s.QueueType = QueueType

    # Constants

    s.latency = latency
    assert s.latency >= 0

    # Channel queue placement

    if s.latency == 0:

      s.recv //= s.send

    else:

      s.queues = [ QueueType( num_entries=2 ) for i in range( s.latency ) ]

      s.recv //= s.queues[0].enq

      @s.update
      def chnl_up_send():

        for i in range( s.latency-1 ):
          if s.queues[i].deq.rdy() and s.queues[i+1].enq.rdy():
            s.queues[i+1].enq( s.queues[i].deq() )

        if s.queues[-1].deq.rdy() and s.send.rdy():
          s.send( s.queues[-1].deq() )

  def line_trace( s ):
    return f"{s.recv}(){s.send}"



'''
==========================================================================
src_sink_test.py
==========================================================================
Unit tests for sources and sinks.

Author : Yanghui Ou
  Date : Feb 3, 2019
'''
from pymtl3 import *
from ..packets   import MflitPacket
from .test_srcs  import MflitPacketSourceCL, MflitPacketSourceRTL
from .test_sinks import MflitPacketSinkCL, MflitPacketSinkRTL

@bitstruct
class SimpleFormat:
  opq  : Bits8
  src  : Bits8
  dst  : Bits8
  plen : Bits8

def mk_simple_pkt( src, dst, payload ):
  plen          = len(payload)
  header        = b32(0)
  header[0 :8 ] = b8(plen)
  header[8 :16] = b8(src)
  header[16:24] = b8(dst)
  flits = [ header ]
  flits.extend( [ b32(x) for x in payload ] )
  return MflitPacket( SimpleFormat, flits )

class TestHarness( Component ):

  def construct( s, Format, pkts ):
    s.src  = MflitPacketSourceCL( SimpleFormat, pkts )
    s.sink = MflitPacketSinkCL  ( SimpleFormat, pkts )
    connect( s.src.send, s.sink.recv )

  def done( s ):
    return s.src.done() and s.sink.done()

  def line_trace( s ):
    return f'{s.src.line_trace()} >>> {s.sink.line_trace()}'

class TestHarnessRTL( Component ):

  def construct( s, Format, pkts ):
    s.src  = MflitPacketSourceRTL( SimpleFormat, pkts )
    s.sink = MflitPacketSinkRTL  ( SimpleFormat, pkts )
    connect( s.src.send, s.sink.recv )

  def done( s ):
    return s.src.done() and s.sink.done()

  def line_trace( s ):
    return f'{s.src.line_trace()} >>> {s.sink.line_trace()}'

#-------------------------------------------------------------------------
# test_simple
#-------------------------------------------------------------------------

def test_simple():

  pkts = [
    mk_simple_pkt( 0, 1, [ 0xbeef, 0xb00c ] ),
    mk_simple_pkt( 1, 0, [ 0xf00d, 0xbabe ] ),
  ]
  th = TestHarness( SimpleFormat, pkts )
  th.elaborate()
  th.apply( SimulationPass() )

  print()
  for i in range(10):
    th.tick()
    print( f'{i:3}:{th.line_trace()}' )

  assert th.done()

#-------------------------------------------------------------------------
# test_src_delay
#-------------------------------------------------------------------------

def test_src_delay():
  pkts = [
    mk_simple_pkt( 0, 1, [ 0xbeef, 0xface, 0xf00d ] ),
    mk_simple_pkt( 0, 1, [ 0xbaad, 0xc0de         ] ),
  ]

  th = TestHarness( SimpleFormat, pkts )
  th.set_param( 'top.src.construct', initial_delay=5, flit_interval_delay=2 )
  th.elaborate()
  th.apply( SimulationPass() )
  print()
  for i in range( 25 ):
    th.tick()
    print( f'{i:3}:{th.line_trace()}' )
  assert th.done()

#-------------------------------------------------------------------------
# test_sink_delay
#-------------------------------------------------------------------------

def test_sink_delay():
  pkts = [
    mk_simple_pkt( 0, 1, [ 0xbeef, 0xface, 0xf00d ] ),
    mk_simple_pkt( 0, 1, [ 0xbaad, 0xc0de         ] ),
  ]

  th = TestHarness( SimpleFormat, pkts )
  th.set_param( 'top.sink.construct', initial_delay=5, flit_interval_delay=2 )
  th.elaborate()
  th.apply( SimulationPass() )
  print()
  for i in range( 25 ):
    th.tick()
    print( f'{i:3}:{th.line_trace()}' )
  assert th.done()

#-------------------------------------------------------------------------
# test_mix_delay
#-------------------------------------------------------------------------

def test_mix_delay():
  pkts = [
    mk_simple_pkt( 0, 1, [ 0xbeef, 0xface, 0xf00d ] ),
    mk_simple_pkt( 0, 1, [ 0xbaad, 0xc0de         ] ),
    mk_simple_pkt( 1, 1, [                        ] ),
    mk_simple_pkt( 2, 1, [ 0xcba                  ] ),
  ]

  th = TestHarness( SimpleFormat, pkts )
  th.set_param( 'top.sink.construct', initial_delay=5, flit_interval_delay=2 )
  th.set_param( 'top.src.construct',  packet_interval_delay=5 )
  th.elaborate()
  th.apply( SimulationPass() )
  print()
  for i in range( 50 ):
    th.tick()
    print( f'{i:3}:{th.line_trace()}' )
  assert th.done()

#-------------------------------------------------------------------------
# test_simple_rtl
#-------------------------------------------------------------------------

def test_simple_rtl():

  pkts = [
    mk_simple_pkt( 0, 1, [ 0xbeef, 0xb00c ] ),
    mk_simple_pkt( 1, 0, [ 0xf00d, 0xbabe ] ),
  ]
  th = TestHarnessRTL( SimpleFormat, pkts )
  th.elaborate()
  th.apply( SimulationPass() )

  print()
  for i in range(10):
    th.tick()
    print( f'{i:3}:{th.line_trace()}' )

  assert th.done()



"""
==========================================================================
RingRouteUnitRTL.py
==========================================================================
A ring route unit with get/give interface.

Author : Yanghui Ou, Cheng Tan
  Date : April 6, 2019
"""
from copy import deepcopy

from pymtl3 import *
from pymtl3.stdlib.ifcs import GetIfcRTL, GiveIfcRTL, SendIfcRTL

from .directions import *


class RingRouteUnitRTL( Component ):

  def construct( s, PacketType, PositionType, num_routers=4 ):

    # Constants
    s.num_outports = 3
    s.num_routers  = num_routers

    id_type    = mk_bits( clog2( num_routers ) )
    dist_type  = mk_bits( clog2( num_routers+1 ) )
    out_type   = mk_bits( s.num_outports )
    dir_type   = mk_bits( clog2( s.num_outports ) )
    s.last_idx = dist_type( num_routers-1 )

    # Interface

    s.get  = GetIfcRTL( PacketType )
    s.give = [ GiveIfcRTL (PacketType) for _ in range ( s.num_outports ) ]
    s.pos  = InPort( PositionType )

    # Componets

    s.out_dir  = Wire( mk_bits( clog2( s.num_outports ) ) )
    s.give_ens = Wire( mk_bits( s.num_outports ) )

    s.left_dist  = Wire( mk_bits( clog2(s.num_routers+1) ) )
    s.right_dist = Wire( mk_bits( clog2(s.num_routers+1) ) )
    s.give_msg_wire = Wire( PacketType )

    # Connections

    for i in range( s.num_outports ):
      s.give_ens[i] //= s.give[i].en

    # Routing logic
    @s.update
    def up_left_right_dist():
      if s.get.ret.dst < s.pos:
        s.left_dist  = dist_type(s.pos) - dist_type(s.get.ret.dst)
        s.right_dist = s.last_idx - dist_type(s.pos) + dist_type(s.get.ret.dst) + dist_type(1)
      else:
        s.left_dist  = dist_type(1) + s.last_idx + dist_type(s.pos) - dist_type(s.get.ret.dst)
        s.right_dist = dist_type(s.get.ret.dst) - dist_type(s.pos)

    @s.update
    def up_ru_routing():

      s.out_dir = dir_type(0)
      s.give_msg_wire = deepcopy( s.get.ret )
      for i in range( s.num_outports ):
        s.give[i].rdy = b1(0)

      if s.get.rdy:
        if s.pos == s.get.ret.dst:
          s.out_dir = dir_type(SELF)
        elif s.left_dist < s.right_dist:
          s.out_dir = dir_type(LEFT)
        else:
          s.out_dir = dir_type(RIGHT)

        if dist_type(s.pos) == s.last_idx and s.out_dir == dir_type(RIGHT):
          s.give_msg_wire.vc_id = b1(1)
        elif s.pos == id_type(0) and s.out_dir == dir_type(LEFT):
          s.give_msg_wire.vc_id = b1(1)

        s.give[ s.out_dir ].rdy = b1(1)
        s.give[ s.out_dir ].ret = s.give_msg_wire

    @s.update
    def up_ru_get_en():
      s.get.en = s.give_ens > out_type(0)

  # Line trace
  def line_trace( s ):

    out_str = [ "" for _ in range( s.num_outports ) ]
    for i in range (s.num_outports):
      out_str[i] = "{}".format( s.give[i] )

    return "{}({}){}".format( s.get, s.out_dir, "|".join( out_str ) )



"""
==========================================================================
OutputUnitRTL_test.py
==========================================================================
Test cases for OutputUnitRTL.

Author : Yanghui Ou, Cheng Tan
  Date : June 22, 2019
"""
import pytest

from pymtl3 import *
from pymtl3.stdlib.rtl.queues import (BypassQueueRTL, NormalQueueRTL,
                                      PipeQueueRTL)
from pymtl3.stdlib.test.test_sinks import TestSinkCL
from pymtl3.stdlib.test.test_srcs import TestSrcCL
from router.OutputUnitRTL import OutputUnitRTL

from .OutputUnitCL_test import OutputUnitCL_Tests as BaseTests

#-------------------------------------------------------------------------
# TestHarness
#-------------------------------------------------------------------------

class TestHarness( Component ):

  def construct( s, MsgType, src_msgs, sink_msgs ):

    s.src   = TestSrcCL( MsgType, src_msgs )
    s.src_q = BypassQueueRTL( MsgType, num_entries=1 )
    s.dut   = OutputUnitRTL( MsgType )
    s.sink  = TestSinkCL( MsgType, sink_msgs )

    # Connections
    s.src.send  //= s.src_q.enq
    s.src_q.deq //= s.dut.get
    s.dut.send  //= s.sink.recv

  def done( s ):
    return s.src.done() and s.sink.done()

  def line_trace( s ):
    return "{}>{}>{}".format(
      s.src.line_trace(),
      s.dut.line_trace(),
      s.sink.line_trace()
    )

#-------------------------------------------------------------------------
# Test cases
#-------------------------------------------------------------------------


class OutputUnitRTL_Tests( BaseTests ):

  @classmethod
  def setup_class( cls ):
    cls.TestHarness = TestHarness
    cls.qtypes      = [ NormalQueueRTL, PipeQueueRTL, BypassQueueRTL ]



"""
==========================================================================
MeshNetworkCL_test.py
==========================================================================
Test for NetworkCL

Author : Yanghui Ou
  Date : May 19, 2019
"""
import pytest

from meshnet.MeshNetworkCL import MeshNetworkCL
from ocnlib.ifcs.packets import mk_mesh_pkt
from ocnlib.ifcs.positions import mk_mesh_pos
from ocnlib.utils import run_sim
from ocnlib.test.net_sinks import TestNetSinkCL
from pymtl3 import *
from pymtl3.stdlib.test import mk_test_case_table
from pymtl3.stdlib.test.test_srcs import TestSrcCL
from router.InputUnitCL import InputUnitCL

#-------------------------------------------------------------------------
# TestHarness
#-------------------------------------------------------------------------

class TestHarness( Component ):

  def construct( s, PktType, ncols, nrows,
                 src_msgs, sink_msgs,
                 src_initial, src_interval,
                 sink_initial, sink_interval ):

    s.nrouters = ncols * nrows

    MeshPos = mk_mesh_pos( ncols, nrows )
    match_func = lambda a, b : a==b
    s.dut = MeshNetworkCL( PktType, MeshPos, ncols, nrows, 0 )

    s.srcs  = [ TestSrcCL( PktType, src_msgs[i],  src_initial,  src_interval  )
                for i in range( s.nrouters ) ]
    s.sinks = [ TestNetSinkCL( PktType, sink_msgs[i], sink_initial, sink_interval, match_func=match_func)
                for i in range( s.nrouters ) ]

    # Connections
    for i in range ( s.nrouters ):
      s.srcs[i].send //= s.dut.recv[i]
      s.dut.send[i]  //= s.sinks[i].recv

  def done( s ):
    srcs_done = 1
    sinks_done = 1
    for i in range( s.nrouters ):
      if s.srcs[i].done() == 0:
        srcs_done = 0
    for i in range( s.nrouters ):
      if s.sinks[i].done() == 0:
        sinks_done = 0
    return srcs_done and sinks_done

  def line_trace( s ):
    return s.dut.line_trace()

#-------------------------------------------------------------------------
# Helper functions
#-------------------------------------------------------------------------

def mk_src_sink_msgs( PktType, msgs, ncols, nrows ):
  nrouters = ncols * nrows
  src_msgs  = [ [] for _ in range( nrouters ) ]
  sink_msgs = [ [] for _ in range( nrouters ) ]

  for msg in msgs:
    src_x, src_y, dst_x, dst_y, opq, payload = msg
    src_id  = src_y * ncols + src_x
    sink_id = dst_y * ncols + dst_x

    src_msgs [ src_id ] .append( PktType(*msg) )
    sink_msgs[ sink_id ].append( PktType(*msg) )

  return src_msgs, sink_msgs

def mk_pkt_list( PktType, lst ):
  ret = []
  for m in lst:
    src_x, src_y, dst_x, dst_y, opq, payload = m[0], m[1], m[2], m[3], m[4], m[5]
    ret.append( PktType( src_x, src_y, dst_x, dst_y, opq, payload ) )
  return ret

#-------------------------------------------------------------------------
# Test cases
#-------------------------------------------------------------------------

simple_2x2 = [
#   src_x src_y dst_x dst_y opq   payload
  ( 0,    0,    0,    1,    0x00, 0x0010  ),
  ( 1,    0,    1,    1,    0x01, 0x0020  ),
]

simple_4x4 = [
#   src_x src_y dst_x dst_y opq   payload
  ( 0,    0,    0,    1,    0x00, 0x0010  ),
  ( 1,    0,    1,    1,    0x01, 0x0020  ),
  ( 3,    2,    1,    1,    0x02, 0x0020  ),
  ( 1,    0,    1,    1,    0x03, 0x0020  ),
  ( 1,    3,    2,    1,    0x04, 0x0020  ),
  ( 3,    3,    1,    0,    0x05, 0x0020  ),
]

simple_8x8 = [
#   src_x src_y dst_x dst_y opq   payload
  ( 0,    0,    0,    1,    0x00, 0x0010  ),
  ( 1,    0,    1,    1,    0x01, 0x0020  ),
  ( 3,    2,    1,    1,    0x02, 0x0020  ),
  ( 1,    0,    1,    1,    0x03, 0x0020  ),
  ( 1,    3,    2,    1,    0x04, 0x0020  ),
  ( 3,    5,    1,    0,    0x05, 0x0020  ),
]

#-------------------------------------------------------------------------
# test case table
#-------------------------------------------------------------------------

test_case_table = mk_test_case_table([
  (            "msg_list    wid  ht  src_init src_intv sink_init sink_intv"),
  ["simple2x2", simple_2x2, 2,   2,  0,       0,       0,        0         ],
  ["simple4x4", simple_4x4, 4,   4,  0,       0,       0,        0         ],
  ["simple8x8", simple_8x8, 8,   8,  0,       0,       0,        0         ],
])

#-------------------------------------------------------------------------
# run test
#-------------------------------------------------------------------------

@pytest.mark.parametrize( **test_case_table )
def test_mesh_simple( test_params ):
  PktType = mk_mesh_pkt( ncols=test_params.wid,
                         nrows=test_params.ht, vc=1 )

  src_msgs, sink_msgs = mk_src_sink_msgs( PktType, test_params.msg_list,
                                          test_params.wid, test_params.ht )
  th = TestHarness(
    PktType,
    test_params.wid,
    test_params.ht,
    src_msgs,
    sink_msgs,
    test_params.src_init,
    test_params.src_intv,
    test_params.sink_init,
    test_params.sink_intv,
  )
  run_sim( th )



"""
=========================================================================
InputUnitGiveRTL.py
=========================================================================
An input unit with a recv interface as input and give interface as output.

Author : Yanghui Ou, Cheng Tan
  Date : Mar 23, 2019
"""
from pymtl3 import *
from pymtl3.stdlib.ifcs import GiveIfcRTL, RecvIfcRTL
from pymtl3.stdlib.rtl.queues import NormalQueueRTL


class InputUnitRTL( Component ):

  def construct( s, PacketType, QueueType = NormalQueueRTL ):

    # Interface

    s.recv = RecvIfcRTL( PacketType )
    s.give = GiveIfcRTL( PacketType )

    # Component

    s.queue = QueueType( PacketType )( enq = s.recv, deq = s.give )

  def line_trace( s ):
    return f"{s.recv}({s.queue.count}){s.give}"



#=========================================================================
# InputUnitRTLSourceSink_test.py
#=========================================================================
# Test for InputUnitRTL using Source and Sink
#
# Author : Cheng Tan, Yanghui Ou
#   Date : Feb 23, 2019

import pytest

from channel.ChannelRTL import ChannelRTL
from ocnlib.utils import run_sim
from pymtl3 import *
from pymtl3.stdlib.rtl.queues import NormalQueueRTL
from pymtl3.stdlib.test import TestVectorSimulator
from pymtl3.stdlib.test.test_sinks import TestSinkRTL
from pymtl3.stdlib.test.test_srcs import TestSrcRTL

#-------------------------------------------------------------------------
# TestHarness
#-------------------------------------------------------------------------

class TestHarness( Component ):

  def construct( s, MsgType, src_msgs, sink_msgs ):

    s.src  = TestSrcRTL ( MsgType, src_msgs  )
    s.sink = TestSinkRTL( MsgType, sink_msgs )
    s.dut  = ChannelRTL( MsgType )

    # Connections
    s.src.send //= s.dut.recv
    s.dut.send //= s.sink.recv

  def done( s ):
    return s.src.done() and s.sink.done()

  def line_trace( s ):
    return s.src.line_trace() + "-> | " + s.dut.line_trace() + \
                               " | -> " + s.sink.line_trace()

#-------------------------------------------------------------------------
# Test cases
#-------------------------------------------------------------------------

test_msgs = [ b16(4), b16(1), b16(2), b16(3) ]

def test_normal2_simple():
  th = TestHarness( Bits16, test_msgs, test_msgs)
  run_sim( th )



"""
==========================================================================
InputUnitCreditRTL.py
==========================================================================
An input unit with a credit based interface.

Author : Yanghui Ou
  Date : June 22, 2019
"""
from ocnlib.ifcs.CreditIfc import CreditRecvIfcRTL
from pymtl3 import *
from pymtl3.stdlib.ifcs import GiveIfcRTL
from pymtl3.stdlib.rtl.arbiters import RoundRobinArbiterEn
from pymtl3.stdlib.rtl.Encoder import Encoder
from pymtl3.stdlib.rtl.queues import NormalQueueRTL


class InputUnitCreditRTL( Component ):

  def construct( s, PacketType, QueueType = NormalQueueRTL,
                 vc=2, credit_line=2 ):

    # Local paramters
    s.vc = vc

    # Interface
    s.recv = CreditRecvIfcRTL( PacketType, vc=vc )
    s.give = [ GiveIfcRTL( PacketType ) for _ in range( vc ) ]

    # Component
    VcIDType = mk_bits( clog2( vc ) ) if vc > 1 else Bits1

    s.buffers = [ QueueType( PacketType, num_entries=credit_line )
                  for _ in range( vc ) ]

    for i in range( vc ):
      s.buffers[i].enq.msg //= s.recv.msg
      s.buffers[i].deq     //= s.give[i]
      s.recv.yum[i]        //= s.give[i].en

    @s.update
    def up_enq():
      if s.recv.en:
        for i in range( vc ):
          s.buffers[i].enq.en = s.recv.msg.vc_id == VcIDType(i)
      else:
        for i in range( vc ):
          s.buffers[i].enq.en = b1(0)

  def line_trace( s ):
    return "{}({}){}".format(
      s.recv,
      ",".join([ str(s.buffers[i].count) for i in range( s.vc ) ]),
      "|".join([ str(s.give[i]) for i in range( s.vc ) ]),
    )



#=========================================================================
# TorusRouterCL.py
#=========================================================================
# Torus network-on-chip router in CL modeling
#
# Author : Cheng Tan
#   Date : May 20, 2019

from DORYTorusRouteUnitCL import DORYTorusRouteUnitCL
from pymtl3 import *
from router.InputUnitCL import InputUnitCL
from router.OutputUnitCL import OutputUnitCL
from router.Router import Router
from router.SwitchUnitCL import SwitchUnitCL


class TorusRouterCL( Router ):

  def construct( s,
                 PacketType,
                 PositionType,
                 InputUnitType  = InputUnitCL,
                 RouteUnitType  = DORYTorusRouteUnitCL,
                 SwitchUnitType = SwitchUnitCL,
                 OutputUnitType = OutputUnitCL,
                 #RecvIfcType = GuardedCalleeIfc,
                 #SendIfcType = GuardedCallerIfc,
                 ):

    s.num_inports  = 5
    s.num_outports = 5

    # Interface
    s.pos  = InPort( PositionType )
    s.recv = [ CalleeIfcCL() for _ in range( s.num_inports  ) ]
    s.send = [ CallerIfcCL() for _ in range( s.num_outports ) ]

    # Components

    s.input_units  = [ InputUnitType( PacketType )
                      for _ in range( s.num_inports ) ]

    s.route_units  = [ RouteUnitType( PacketType, PositionType, s.num_outports )
                      for i in range( s.num_inports ) ]

    s.switch_units = [ SwitchUnitType( PacketType, s.num_inports )
                      for _ in range( s.num_outports ) ]

    s.output_units = [ OutputUnitType( PacketType )
                      for _ in range( s.num_outports ) ]

    # Connection

    for i in range( s.num_inports ):
      s.connect( s.recv[i],             s.input_units[i].recv )
      s.connect( s.input_units[i].give, s.route_units[i].get  )
      s.connect( s.pos,                 s.route_units[i].pos  )

    for i in range( s.num_inports ):
      for j in range( s.num_outports ):
        s.connect( s.route_units[i].give[j], s.switch_units[j].get[i] )

    for j in range( s.num_outports ):
      s.connect( s.switch_units[j].send, s.output_units[j].recv )
      s.connect( s.output_units[j].send, s.send[j]              )

  def line_trace( s ):
    return "{}>{}".format(
      "|".join( [ s.input_units[i].line_trace()  for i in range(5) ] ),
      "|".join( [ s.route_units[i].line_trace() for i in range(5) ] ),
      #"|".join( [ s.switch_units[i].line_trace() for i in range(3) ] ),
    )



'''
==========================================================================
MeshRouterMflitRTL
==========================================================================
Mesh router that supports multi-flit packet.

Author : Yanghui Ou
  Date : Feb 13, 2020
'''
from pymtl3 import *
from pymtl3.stdlib.ifcs import RecvIfcRTL, SendIfcRTL
from ocnlib.utils import get_nbits
from router.InputUnitRTL import InputUnitRTL
from router.OutputUnitRTL import OutputUnitRTL
from router.SwitchUnitGrantHoldRTL import SwitchUnitGrantHoldRTL

from .MeshRouteUnitRTLMflitXY import MeshRouteUnitRTLMflitXY

#-------------------------------------------------------------------------
# MeshRouterMflitRTL
#-------------------------------------------------------------------------

class MeshRouterMflitRTL( Component ):

  #-----------------------------------------------------------------------
  # Construct
  #-----------------------------------------------------------------------

  def construct( s,
    Header,
    PositionType,
    InputUnitType  = InputUnitRTL,
    RouteUnitType  = MeshRouteUnitRTLMflitXY,
    SwitchUnitType = SwitchUnitGrantHoldRTL,
    OutputUnitType = OutputUnitRTL,
  ):

    # Local parameter

    s.num_inports  = 5
    s.num_outports = 5
    s.PhitType     = mk_bits( get_nbits( Header ) )

    # Interface

    s.recv = [ RecvIfcRTL( s.PhitType ) for _ in range( s.num_inports  ) ]
    s.send = [ SendIfcRTL( s.PhitType ) for _ in range( s.num_outports ) ]
    s.pos  = InPort( PositionType )

    # Components

    s.input_units  = [ InputUnitType( s.PhitType )
                       for _ in range( s.num_inports ) ]

    s.route_units  = [ RouteUnitType( Header, PositionType )
                       for i in range( s.num_inports ) ]

    s.switch_units = [ SwitchUnitType( s.PhitType, s.num_inports )
                       for _ in range( s.num_outports ) ]

    s.output_units = [ OutputUnitType( s.PhitType )
                       for _ in range( s.num_outports ) ]

    # Connections

    for i in range( s.num_inports ):
      s.recv[i]             //= s.input_units[i].recv
      s.input_units[i].give //= s.route_units[i].get
      s.pos                 //= s.route_units[i].pos

    for i in range( s.num_inports ):
      for j in range( s.num_outports ):
        s.route_units[i].give[j] //= s.switch_units[j].get[i]
        s.route_units[i].hold[j] //= s.switch_units[j].hold[i]

    for j in range( s.num_outports ):
      s.switch_units[j].give //= s.output_units[j].get
      s.output_units[j].send //= s.send[j]

  #-----------------------------------------------------------------------
  # Line trace
  #-----------------------------------------------------------------------

  def line_trace( s ):
    in_trace  = '|'.join([ f'{ifc}' for ifc in s.recv ])
    out_trace = '|'.join([ f'{ifc}' for ifc in s.send ])
    return f'{in_trace}({s.pos}){out_trace}'



"""
==========================================================================
RingNetworkFL
==========================================================================
Functional level of a ring network.

Author : Yanghui Ou
  Date : June 28, 2019
"""

from pymtl3 import *

#-------------------------------------------------------------------------
# ringnet_fl
#-------------------------------------------------------------------------
# [src_pkts] is a list of lists of packets, where each list of packets
# represents traffic from a certain terminal. The FL model returns
# [dst_pkts], which is also a list of lists of packets, where each list of
# packets represents the output packets of each terminal.

def ringnet_fl( src_pkts ):
  nterminals = len( src_pkts )
  dst_pkts = [ [] for _ in range( nterminals ) ]

  dst = 0
  for packets in src_pkts:
    for pkt in packets:
      # TODO: change vc_id?
      if hasattr(pkt, 'fl_type'):
        if pkt.fl_type == 0:
          dst = pkt.dst
      else:
        dst = pkt.dst
      dst_pkts[ dst ].append( pkt )
  return dst_pkts

# class RingNetworkFL( Component ):
#   def construct( s, nterminals ):
#     s.recv = [ RecvIfcFL( method=s.recv_ ) for _ in range( nterminals ) ]
#     s.send = [ SendIfcFL() for _ in range( nterminals ) ]
# 
#     s.out_q = [ QueueFL() for _ in range( nterminals ) ]
# 
#     for i in range( nterminals ):
#       s.connect( s.out_q.send, s.send )
# 
#   def recv_( s, pkt ):
#     s.out_q[ int(pkt.dst) ].enq( pkt )



"""
=========================================================================
RingNetworkCL.py
=========================================================================
Cycle level ring network implementation.

Author : Yanghui Ou
  Date : May 19, 2019
"""
from channel.ChannelCL import ChannelCL
from directions import *
from pymtl3 import *
from RingRouterCL import RingRouterCL


class RingNetworkCL( Component ):
  def construct( s, PacketType, PositionType, nrouters=4, chl_lat=0 ):

    # Constants

    s.nrouters = nrouters
    num_channels  = nrouters * 2
    s.num_terminals = nrouters

    # Interface

    s.recv = [ CalleeIfcCL() for _ in range(s.num_terminals) ]
    s.send = [ CallerIfcCL() for _ in range(s.num_terminals) ]

    # Components

    s.routers    = [ RingRouterCL( PacketType, PositionType )
                     for i in range( s.nrouters ) ]

    s.channels   = [ ChannelCL( PacketType, latency = chl_lat)
                     for _ in range( num_channels ) ]

    # Connect s.routers together in Mesh

    chl_id = 0
    for i in range( s.nrouters ):
      s.connect( s.routers[i].send[RIGHT], s.channels[chl_id].recv                   )
      s.connect( s.channels[chl_id].send,  s.routers[(i+1)%nrouters].recv[LEFT] )
      chl_id += 1

      s.connect( s.routers[(i+1)%nrouters].send[LEFT], s.channels[chl_id].recv  )
      s.connect( s.channels[chl_id].send,                 s.routers[i].recv[RIGHT] )
      chl_id += 1

      # Connect the self port (with Network Interface)

      s.connect(s.recv[i], s.routers[i].recv[SELF])
      s.connect(s.send[i], s.routers[i].send[SELF])

    # FIXME: unable to connect a struct to a port.
    @s.update
    def up_pos():
      for r in range( s.nrouters ):
        s.routers[r].pos = PositionType( r )

  def line_trace( s ):
    return "|".join(
      [ s.routers[i].line_trace() for i in range( s.num_terminals ) ]
    )



'''
==========================================================================
MeshNetworkMflitRTL.py
==========================================================================
Mesh network that supports multi-flit packet.

Author : Yanghui Ou
  Date : Feb 13, 2020
'''
from pymtl3 import *
from pymtl3.stdlib.ifcs import RecvIfcRTL, SendIfcRTL

from .directions import *
from .MeshRouterMflitRTL import MeshRouterMflitRTL


class MeshNetworkMflitRTL( Component ):

  #-----------------------------------------------------------------------
  # construct
  #-----------------------------------------------------------------------
  # TODO: parametrize channel latency.

  def construct( s, Header, PositionType, ncols=2, nrows=2 ):

    # Local parameters

    s.nterminals = ncols * nrows
    s.PhitType   = mk_bits( get_nbits( Header ) )

    # Interface

    s.recv = [ RecvIfcRTL( s.PhitType ) for _ in range( s.nterminals ) ]
    s.send = [ SendIfcRTL( s.PhitType ) for _ in range( s.nterminals ) ]

    # Components

    s.routers = [ MeshRouterMflitRTL( Header, PositionType )
                  for _ in range( s.nterminals ) ]

    # Set position of each router

    for y in range( nrows ):
      for x in range( ncols ):
        s.routers[ y*ncols + x ].pos //= PositionType( x, y )

    # Connect routers

    for i in range( s.nterminals ):
      if i // ncols > 0:
        s.routers[i].send[SOUTH] //= s.routers[i-ncols].recv[NORTH]

      if i // ncols < nrows - 1:
        s.routers[i].send[NORTH] //= s.routers[i+ncols].recv[SOUTH]

      if i % ncols > 0:
        s.routers[i].send[WEST] //= s.routers[i-1].recv[EAST]

      if i % ncols < ncols - 1:
        s.routers[i].send[EAST] //= s.routers[i+1].recv[WEST]

      # Connect the self port (with Network Interface)

      s.recv[i] //= s.routers[i].recv[SELF]
      s.send[i] //= s.routers[i].send[SELF]

      # Connect the unused ports
      # FIXME: for now we hackily ground the payload field so that pymtl
      # won't complain about net need driver.

      if i // ncols == 0:
        s.routers[i].send[SOUTH].rdy //= 0
        s.routers[i].recv[SOUTH].en  //= 0
        s.routers[i].recv[SOUTH].msg //= 0

      if i // ncols == nrows - 1:
        s.routers[i].send[NORTH].rdy //= 0
        s.routers[i].recv[NORTH].en  //= 0
        s.routers[i].recv[NORTH].msg //= 0

      if i % ncols == 0:
        s.routers[i].send[WEST].rdy //= 0
        s.routers[i].recv[WEST].en  //= 0
        s.routers[i].recv[WEST].msg //= 0

      if i % ncols == ncols - 1:
        s.routers[i].send[EAST].rdy //= 0
        s.routers[i].recv[EAST].en  //= 0
        s.routers[i].recv[EAST].msg //= 0

  #-----------------------------------------------------------------------
  # line_trace
  #-----------------------------------------------------------------------

  def line_trace( s ):
    in_trace  = '|'.join([ f'{ifc}' for ifc in s.recv ])
    out_trace = '|'.join([ f'{ifc}' for ifc in s.send ])
    return f'{in_trace}(){out_trace}'



"""
=========================================================================
RingRouteRTL.py
=========================================================================
Ring network-on-chip router

Author : Yanghui Ou, Cheng Tan
  Date : June 25, 2019
"""
from ocnlib.ifcs.CreditIfc import CreditRecvIfcRTL, CreditSendIfcRTL
from pymtl3 import *
from router.InputUnitCreditRTL import InputUnitCreditRTL
from router.OutputUnitCreditRTL import OutputUnitCreditRTL
from router.Router import Router
from router.SwitchUnitRTL import SwitchUnitRTL

from .RingRouteUnitRTL import RingRouteUnitRTL


class RingRouterRTL( Router ):

  def construct( s,
    PacketType,
    PositionType,
    num_routers,
    InputUnitType=InputUnitCreditRTL,
    RouteUnitType=RingRouteUnitRTL,
    SwitchUnitType=SwitchUnitRTL,
    OutputUnitType=OutputUnitCreditRTL,
    vc=2,
    credit_line=2,
  ):

    # Local paramters

    s.num_inports  = 3
    s.num_outports = 3
    s.vc = vc
    s.num_route_units = s.num_inports * s.vc

    # Interface

    s.pos  = InPort( PositionType )
    s.recv = [ CreditRecvIfcRTL( PacketType, s.vc ) for _ in range( s.num_inports  ) ]
    s.send = [ CreditSendIfcRTL( PacketType, s.vc ) for _ in range( s.num_outports ) ]

    # Components

    s.input_units  = [ InputUnitType( PacketType, vc=vc, credit_line=credit_line )
                       for _ in range( s.num_inports ) ]

    s.route_units  = [ RouteUnitType( PacketType, PositionType, num_routers )
                       for i in range( s.num_route_units ) ]

    s.switch_units = [ SwitchUnitType( PacketType, s.num_route_units )
                       for _ in range( s.num_outports ) ]

    s.output_units = [ OutputUnitType( PacketType )
                       for _ in range( s.num_outports ) ]

    # Connection

    for i in range( s.num_inports ):
      s.recv[i] //= s.input_units[i].recv
      for j in range( s.vc ):
        ru_idx = i * s.vc + j
        s.input_units[i].give[j] //= s.route_units[ru_idx].get
        s.pos                    //= s.route_units[ru_idx].pos

    for i in range( s.num_route_units ):
      for j in range( s.num_outports ):
        s.route_units[i].give[j] //= s.switch_units[j].get[i]

    for j in range( s.num_outports ):
      s.switch_units[j].give //= s.output_units[j].get
      s.output_units[j].send //= s.send[j]

  # Line trace

  def line_trace( s ):

    in_trace  = [ "" for _ in range( s.num_inports  ) ]
    out_trace = [ "" for _ in range( s.num_outports ) ]

    for i in range( s.num_inports ):
      in_trace[i]  = "{}".format( s.recv[i].line_trace() )
    for i in range( s.num_outports ):
      out_trace[i] = "{}".format( s.send[i].line_trace() )

    return "{}({}){}".format(
      "|".join( in_trace ),
      s.pos,
      "|".join( out_trace )
    )



"""
=========================================================================
SwitchUnitRTL.py
=========================================================================
A switch unit with GetIfcRTL and SendIfcRTL.

Author : Yanghui Ou, Cheng Tan
  Date : Feb 28, 2019
"""
from pymtl3 import *
from pymtl3.stdlib.ifcs import GetIfcRTL, GiveIfcRTL, SendIfcRTL
from pymtl3.stdlib.rtl import Mux
from pymtl3.stdlib.rtl.arbiters import RoundRobinArbiterEn
from pymtl3.stdlib.rtl.Encoder import Encoder


class SwitchUnitRTL( Component ):

  def construct( s, PacketType, num_inports=5 ):

    # Local parameters

    s.num_inports = num_inports
    s.sel_width   = clog2( num_inports )
    GrantType     = mk_bits( num_inports )
    SelType       = mk_bits( s.sel_width )
    s.set_ocp = 0
    s.clear_ocp = 0

    # Interface

    s.get  = [ GetIfcRTL( PacketType ) for _ in range( s.num_inports ) ]
    s.give = GiveIfcRTL( PacketType )
    s.out_ocp = OutPort( Bits1 )

    # Components

    s.get_en  = [ Wire( Bits1 ) for _ in range( s.num_inports ) ]
    s.get_rdy = [ Wire( Bits1 ) for _ in range( s.num_inports ) ]

    s.arbiter = RoundRobinArbiterEn( num_inports )( en = 1 )
    s.mux = Mux( PacketType, num_inports )( out = s.give.ret )

    s.encoder = Encoder( num_inports, s.sel_width )(
      in_ = s.arbiter.grants,
      out = s.mux.sel
    )

    # Connections

    for i in range( num_inports ):
      s.get[i].rdy //= s.arbiter.reqs[i]
      s.get[i].ret //= s.mux.in_[i]
      s.get[i].en  //= s.get_en[i]
      s.get[i].rdy //= s.get_rdy[i]

    @s.update
    def up_give():
      s.give.rdy = s.arbiter.grants > GrantType(0)

    @s.update
    def up_get_en():
      for i in range( num_inports ):
        s.get_en[i] = s.give.en & ( s.mux.sel==SelType(i) )

  # Line trace

  def line_trace( s ):
    in_trace = [ str(s.get[i]) for i in range( s.num_inports ) ]
    return "{}(){}".format( "|".join( in_trace ), s.give )



"""
==========================================================================
 CreditIfc_test.py
==========================================================================
Test for credit based interfaces.

Author : Yanghui Ou
  Date : June 10, 2019
"""
from ocnlib.utils import run_sim
from ocnlib.test.net_sinks import TestNetSinkRTL
from pymtl3 import *
from pymtl3.stdlib.rtl.queues import NormalQueueRTL
from pymtl3.stdlib.test.test_sinks import TestSinkRTL
from pymtl3.stdlib.test.test_srcs import TestSrcRTL

from .CreditIfc import (CreditRecvIfcRTL, CreditRecvRTL2SendRTL,
                        CreditSendIfcRTL, RecvRTL2CreditSendRTL)
from .packets import mk_generic_pkt

#-------------------------------------------------------------------------
# TestHarness
#-------------------------------------------------------------------------

class TestHarness( Component ):
  def construct( s, Type, src_msgs, sink_msgs, credit_line=2 ):

    s.src = TestSrcRTL( Type, src_msgs )
    s.src_adapter = RecvRTL2CreditSendRTL( Type, vc=2, credit_line=credit_line )
    s.sink_adapter = CreditRecvRTL2SendRTL( Type, vc=2, credit_line=credit_line, QType=NormalQueueRTL )
    s.sink = TestNetSinkRTL( Type, sink_msgs )

    s.src.send          //= s.src_adapter.recv
    s.src_adapter.send  //= s.sink_adapter.recv
    s.sink_adapter.send //= s.sink.recv

  def line_trace( s ):
    return "{} > {} > {} > {}".format(
      s.src.line_trace(),
      s.src_adapter.line_trace(),
      s.sink_adapter.line_trace(),
      s.sink.line_trace()
    )

  def done( s ):
    return s.src.done() and s.sink.done()

#-------------------------------------------------------------------------
# Test cases
#-------------------------------------------------------------------------

def test_simple():
  Pkt = mk_generic_pkt( vc=2 )
  msgs = [
    Pkt( 0, 1, 0x04, 0, 0xdeadbabe ),
    Pkt( 0, 2, 0x02, 1, 0xfaceb00c ),
    Pkt( 0, 3, 0x03, 0, 0xdeadface ),
  ]
  th = TestHarness( Pkt, msgs, msgs )
  run_sim( th )

def test_backpresure():
  Pkt = mk_generic_pkt( vc=2 )
  msgs = [
     # src dst opq vc_id  payload
    Pkt( 0, 1, 0x04, 0, 0xdeadbabe ),
    Pkt( 0, 2, 0x02, 1, 0xfaceb00c ),
    Pkt( 0, 3, 0x03, 0, 0xdeadface ),
    Pkt( 0, 3, 0x03, 1, 0xdeadface ),
    Pkt( 0, 3, 0x03, 1, 0xdeadface ),
    Pkt( 0, 3, 0x03, 0, 0xdeadface ),
    Pkt( 0, 3, 0x03, 0, 0xdeadface ),
  ]
  th = TestHarness( Pkt, msgs, msgs )
  th.set_param( "top.sink.construct", initial_delay=20)
  run_sim( th )



"""
==========================================================================
TorusRouterRTL_test.py
==========================================================================
Tests for TorusRouterRTL.

Author : Yanghui Ou, Cheng Tan
  Date : June 28, 2019
"""
from itertools import product

import pytest

from ocnlib.ifcs.CreditIfc import (CreditRecvRTL2SendRTL,
                                      RecvRTL2CreditSendRTL)
from ocnlib.ifcs.packets import mk_mesh_pkt
from ocnlib.ifcs.positions import mk_mesh_pos
from ocnlib.utils import run_sim
from ocnlib.test.net_sinks import TestNetSinkRTL
from pymtl3 import *
from pymtl3.stdlib.test.test_srcs import TestSrcRTL
from torusnet.TorusRouterFL import TorusRouterFL
from torusnet.TorusRouterRTL import TorusRouterRTL

#-------------------------------------------------------------------------
# TestHarness
#-------------------------------------------------------------------------

class TestHarness( Component ):

  def construct( s,
    PktType   = None,
    src_msgs  = [],
    sink_msgs = [],
    ncols     = 2,
    nrows     = 2 ,
    pos_x     = 0,
    pos_y     = 0,
  ):

    MeshPos = mk_mesh_pos( ncols, nrows )

    match_func = lambda a, b : a.src_x == b.src_x and a.src_y == b.src_y and \
                               a.dst_y == b.dst_y and a.payload == b.payload

    s.srcs  = [ TestSrcRTL( PktType, src_msgs[i] )
                for i in range( 5 ) ]
    s.dut   = TorusRouterRTL( PktType, MeshPos, ncols=ncols, nrows=nrows )
    s.sinks = [ TestNetSinkRTL( PktType, sink_msgs[i], match_func=match_func )
                for i in range( 5 ) ]

    s.src_adapters  = [ RecvRTL2CreditSendRTL( PktType, vc=2 )
                        for _ in range( 5 ) ]
    s.sink_adapters = [ CreditRecvRTL2SendRTL( PktType, vc=2 )
                        for _ in range( 5 ) ]

    # Connections
    for i in range ( s.dut.num_outports ):
      s.srcs[i].send          //= s.src_adapters[i].recv
      s.src_adapters[i].send  //= s.dut.recv[i]
      s.dut.send[i]           //= s.sink_adapters[i].recv
      s.sink_adapters[i].send //= s.sinks[i].recv

    @s.update
    def up_pos():
      s.dut.pos = MeshPos( pos_x, pos_y )

  def done( s ):
    srcs_done = True
    sinks_done = True
    for x in s.srcs:
      if not x.done():
        srcs_done = False
    for x in s.sinks:
      if not x.done():
        sinks_done = False
    return srcs_done and sinks_done

  def line_trace( s ):
    return "{}".format( s.dut.line_trace() )

#-------------------------------------------------------------------------
# mk_srcsink_pkts
#-------------------------------------------------------------------------
# A helper function that puts each packet in [lst] into corresponding
# sources and sinks.

def mk_srcsink_pkts( pos_x, pos_y, ncols, nrows, lst ):
  router = TorusRouterFL( pos_x, pos_y, ncols, nrows, dimension='y' )
  src_pkts  = router.arrange_src_pkts( lst )
  sink_pkts = router.route( src_pkts )
  return src_pkts, sink_pkts

#-------------------------------------------------------------------------
# Test cases
#-------------------------------------------------------------------------

class TorusRouterRTL_Tests:

  @pytest.mark.parametrize(
    'pos_x, pos_y',
    product( [ 0, 1, 2, 3 ], [ 0, 1, 2, 3 ] )
  )
  def test_simple_4x4( s, pos_x, pos_y ):
    ncols = 4
    nrows = 4

    Pkt = mk_mesh_pkt( ncols, nrows, vc=2 )

    src_pkts, sink_pkts = mk_srcsink_pkts( pos_x, pos_y, ncols, nrows,[
      #   src_x  y  dst_x  y  opq  vc  payload
      Pkt(    0, 0,     1, 1,   0,  0, 0xfaceb00c ),
      Pkt(    0, 0,     0, 0,   0,  0, 0xdeaddead ),
      Pkt(    1, 0,     1, 0,   0,  0, 0xdeadface ),
      Pkt(    0, 2,     3, 3,   0,  0, 0xdeadface ),
    ])

    th = TestHarness( Pkt, src_pkts, sink_pkts )
    th.set_param( "top.construct",
      ncols=ncols, nrows=nrows,
      pos_x=pos_x, pos_y=pos_y,
    )
    run_sim( th )

  @pytest.mark.parametrize(
    'pos_x, pos_y',
    product( [ 0, 1, 2, 3, 4, 5 ], [ 0, 1, 2, 3, 4, 5 ] )
  )
  def test_simple_5x5( s, pos_x, pos_y ):
    ncols = 5
    nrows = 5

    Pkt = mk_mesh_pkt( ncols, nrows, vc=2 )

    src_pkts, sink_pkts = mk_srcsink_pkts( pos_x, pos_y, ncols, nrows,[
      #   src_x  y  dst_x  y  opq  vc  payload
      Pkt(    1, 0,     0, 4,   0,  0, 0xfaceb00c ),
    ])

    th = TestHarness( Pkt, src_pkts, sink_pkts )
    th.set_param( "top.construct",
      ncols=ncols, nrows=nrows,
      pos_x=pos_x, pos_y=pos_y,
    )
    run_sim( th )



'''
==========================================================================
MeshRouterMflitRTL_test.py
==========================================================================
Test cases for mesh router that supports multi-flit packet.

Author : Yanghui Ou
  Date : Feb 13, 2020
'''
import pytest
from pymtl3 import *
from pymtl3.stdlib.test import mk_test_case_table
from ocnlib.utils import to_bitstruct, run_sim
from ocnlib.packets import MflitPacket as Packet
from ocnlib.test.test_srcs import MflitPacketSourceRTL as TestSource
from ocnlib.test.test_sinks import MflitPacketSinkRTL as TestSink

from ..MeshRouterFL import MeshRouterMflitFL
from ..MeshRouterMflitRTL import MeshRouterMflitRTL

#-------------------------------------------------------------------------
# bitstructs for testing
#-------------------------------------------------------------------------

@bitstruct
class TestHeader:
  opaque : Bits12
  src_x  : Bits4
  src_y  : Bits4
  dst_x  : Bits4
  dst_y  : Bits4
  plen   : Bits4

@bitstruct
class TestPosition:
  pos_x : Bits4
  pos_y : Bits4

#-------------------------------------------------------------------------
# TestHarness
#-------------------------------------------------------------------------

class TestHarness( Component ):

  def construct( s, Header, Position, pos_x, pos_y, src_pkts, sink_pkts ):
    nports = 5

    s.src  = [ TestSource( Header, src_pkts[i] ) for i in range( nports ) ]
    s.dut  = MeshRouterMflitRTL( Header, Position )
    s.sink = [ TestSink( Header, sink_pkts[i] ) for i in range( nports ) ]

    for i in range( nports ):
      s.src[i].send //= s.dut.recv[i]
      s.dut.send[i] //= s.sink[i].recv

    s.dut.pos //= Position( pos_x, pos_y )

  def done( s ):
    src_done = True
    sink_done = True
    for m in s.src:
      src_done &= m.done()
    for m in s.sink:
      sink_done &= m.done()
    return src_done and sink_done

  def line_trace( s ):
    return s.dut.line_trace()

#-------------------------------------------------------------------------
# mk_pkt
#-------------------------------------------------------------------------

def mk_pkt( src_x, src_y, dst_x, dst_y, payload=[], opaque=0 ):
  plen        = len( payload )
  header      = TestHeader( opaque, src_x, src_y, dst_x, dst_y, plen )
  header_bits = to_bits( header )
  flits       = [ header_bits ] + payload
  return Packet( TestHeader, flits )

#-------------------------------------------------------------------------
# sanity check
#-------------------------------------------------------------------------

def test_sanity_check():
  dut = MeshRouterMflitRTL( TestHeader, TestPosition )
  dut.elaborate()
  dut.apply( SimulationPass() )
  dut.sim_reset()
  dut.tick()

#-------------------------------------------------------------------------
# test case: basic
#-------------------------------------------------------------------------

def basic_pkts( pos_x, pos_y ):
  return [
    #       src_x  y      dst_x  y      payload
    mk_pkt(     0, 0,     pos_x, pos_y, [ 0x8badf00d             ] ),
    mk_pkt( pos_x, pos_y, 0,     0,     [ 0x8badf00d, 0xdeadbeef ] ),
  ]

#-------------------------------------------------------------------------
# test case: hotspot
#-------------------------------------------------------------------------

def hotspot_pkts( pos_x, pos_y ):
  return [
    #       src_x  y      dst_x  y      payload
    mk_pkt(     0, 0,     pos_x, pos_y, [ 0x8badf00d                         ] ),
    mk_pkt(     0, 1,     pos_x, pos_y, [ 0x8badf00d, 0xdeadbeef             ] ),
    mk_pkt( pos_x, 0,     pos_x, pos_y, [                                    ] ),
    mk_pkt(     2, pos_y, pos_x, pos_y, [ 0xdeadbeef, 0xfaceb00c, 0xdeadfa11 ] ),
    mk_pkt(     2, pos_y, pos_x, pos_y, [                                    ] ),
    mk_pkt(     2, pos_y, pos_x, pos_y, [ 0xdeadc0de ] * 10                   ),
  ]

#-------------------------------------------------------------------------
# test case: long
#-------------------------------------------------------------------------

def long_pkts( pos_x, pos_y ):
  return [
    #       src_x  y      dst_x  y      payload
    mk_pkt(     0, 0,     pos_x, pos_y, [ 0x8badf00d                         ] * 13 ),
    mk_pkt(     0, 1,     pos_x, pos_y, [ 0x8badf00d, 0xdeadbeef             ] * 6  ),
    mk_pkt( pos_x, 0,     pos_x, pos_y, [                                    ]      ),
    mk_pkt(     2, 1,     pos_x, pos_y, [ 0xdeadbeef, 0xfaceb00c, 0xdeadfa11 ] * 3  ),
    mk_pkt(     7, pos_y, pos_x, pos_y, [                                    ]      ),
    mk_pkt(     0, 8,     pos_x, pos_y, [ 0xdeadc0de                         ] * 10 ),
  ]

#-------------------------------------------------------------------------
# test case table
#-------------------------------------------------------------------------

table = [
  (               'msg_func      pos_x  pos_y' ),
  [ 'basic',       basic_pkts,       1,     1  ],
  [ 'hotspot2_3',  hotspot_pkts,     2,     3  ],
  [ 'hotspot4_4',  hotspot_pkts,     4,     4  ],
  [ 'long_4_4',    long_pkts,        4,     4  ],
]

for x in range(4):
  for y in range(4):
    table.append([ f'hotspot_{x},{y}', hotspot_pkts, x, y ])

for x in range(4):
  for y in range(4):
    table.append([ f'long_{x},{y}', long_pkts, x, y ])

test_case_table = mk_test_case_table( table )

#-------------------------------------------------------------------------
# test driver
#-------------------------------------------------------------------------

@pytest.mark.parametrize( **test_case_table )
def test_mflit_mesh_router( test_params ):
  ref  = MeshRouterMflitFL( TestHeader, test_params.pos_x, test_params.pos_y )
  pkts = test_params.msg_func( test_params.pos_x, test_params.pos_y )

  src_pkts = ref.arrange_src_pkts( pkts )
  dst_pkts = ref.route( src_pkts )

  th = TestHarness( TestHeader, TestPosition,
                    test_params.pos_x, test_params.pos_y,
                    src_pkts, dst_pkts )
  run_sim( th )



'''
==========================================================================
XbarRouteUnitMflitRTL.py
==========================================================================
Route unit for crossbar that supports multi-flit packet.

- NOTE:

It requires no tail flit and that the header flit should have a field that
indicates the length of the payload in terms of number of flits.

Authour : Yanghui Ou
   Date : Feb 18, 2020
'''
from pymtl3 import *
from pymtl3.stdlib.ifcs import GetIfcRTL, GiveIfcRTL
from ocnlib.rtl import Counter, GrantHoldArbiter
from ocnlib.utils import get_nbits, get_plen_type
from ocnlib.utils.connects import connect_bitstruct

class XbarRouteUnitMflitRTL( Component ):

  #-----------------------------------------------------------------------
  # construct
  #-----------------------------------------------------------------------

  def construct( s, HeaderFormat, num_outports=4, plen_field_name='plen' ):
    # Meta data
    s.num_outports = num_outports
    s.HeaderFormat = HeaderFormat
    s.PhitType     = mk_bits( get_nbits( HeaderFormat ) )
    dir_nbits      = clog2( num_outports ) if num_outports > 1 else 1
    DirType        = mk_bits( dir_nbits )
    s.STATE_HEADER = b1(0)
    s.STATE_BODY   = b1(1)

    PLenType = get_plen_type( HeaderFormat )

    # Interface
    s.get  = GetIfcRTL( s.PhitType )

    s.give = [ GiveIfcRTL( s.PhitType ) for _ in range( s.num_outports ) ]
    s.hold = [ OutPort( Bits1 ) for _ in range( s.num_outports ) ]

    # Components
    s.header      = Wire( HeaderFormat )
    s.state       = Wire( Bits1 )
    s.state_next  = Wire( Bits1 )
    s.out_dir_r   = Wire( DirType )
    s.out_dir     = Wire( DirType )
    s.any_give_en = Wire( Bits1 )

    s.counter = Counter( PLenType )(
      incr=b1(0),
      load_value=s.header.plen,
    )

    connect_bitstruct( s.get.ret, s.header )

    for i in range( s.num_outports ):
      s.get.ret //= s.give[i].ret
    s.get.en //= s.any_give_en

    @s.update
    def up_any_give_en():
      s.any_give_en = b1(0)
      for i in range( s.num_outports ):
        if s.give[i].en:
          s.any_give_en = b1(1)

    # State transition logic
    @s.update_ff
    def up_state_r():
      if s.reset:
        s.state <<= s.STATE_HEADER
      else:
        s.state <<= s.state_next

    @s.update
    def up_state_next():
      if s.state == s.STATE_HEADER:
        # If the packet has body flits
        if s.any_give_en & ( s.header.plen > PLenType(0) ):
          s.state_next = s.STATE_BODY

        else:
          s.state_next = s.STATE_HEADER

      else: # STATE_BODY
        if ( s.counter.count == PLenType(1) ) & s.any_give_en:
          s.state_next = s.STATE_HEADER
        else:
          s.state_next = s.STATE_BODY

    # State output logic
    @s.update
    def up_counter_decr():
      if s.state == s.STATE_HEADER:
        s.counter.decr = b1(0)
      else:
        s.counter.decr = s.any_give_en

    @s.update
    def up_counter_load():
      if s.state == s.STATE_HEADER:
        s.counter.load = ( s.state_next == s.STATE_BODY )
      else:
        s.counter.load = b1(0)

    # Routing logic
    # TODO: Figure out how to encode dest id
    @s.update
    def up_out_dir():
      if ( s.state == s.STATE_HEADER ) & s.get.rdy:
        s.out_dir = s.header.dst[0:dir_nbits]

      else:
        s.out_dir = s.out_dir_r

    @s.update_ff
    def up_out_dir_r():
      s.out_dir_r <<= s.out_dir

    @s.update
    def up_give_rdy_hold():
      for i in range( s.num_outports ):
        s.give[i].rdy = ( DirType(i) == s.out_dir ) & s.get.rdy
        s.hold[i]     = ( DirType(i) == s.out_dir ) & ( s.state == s.STATE_BODY )

  #-----------------------------------------------------------------------
  # line_trace
  #-----------------------------------------------------------------------

  def line_trace( s ):
    give_trace = '|'.join([ f'{ifc}' for ifc in s.give ])
    hold  = ''.join([ '^' if h else '.' for h in s.hold ])
    count = f'{s.counter.count}'
    state = 'H' if s.state == s.STATE_HEADER else \
            'B' if s.state == s.STATE_BODY else   \
            '!'
    dir   = f'{s.out_dir}'
    return f'{s.get}([{state}{count}]{dir}{hold}){give_trace}'



"""
=========================================================================
MeshNetworkRTL.py
=========================================================================
Mesh network implementation.

Author : Yanghui Ou, Cheng Tan
  Date : Mar 10, 2019
"""
from channel.ChannelRTL import ChannelRTL
from pymtl3 import *
from pymtl3.stdlib.ifcs.SendRecvIfc import *

from .directions import *
from .MeshRouterRTL import MeshRouterRTL


class MeshNetworkRTL( Component ):
  def construct( s, PacketType, PositionType,
                 ncols=4, nrows=4, chl_lat=0 ):

    # Local parameters

    s.num_routers   = ncols * nrows
    s.num_terminals = s.num_routers
    num_channels    = (nrows*(ncols-1)+ncols*(nrows-1)) * 2
    XType           = mk_bits( clog2(ncols) )
    YType           = mk_bits( clog2(nrows ) )

    # Interface

    s.recv = [ RecvIfcRTL( PacketType ) for _ in range( s.num_terminals )]
    s.send = [ SendIfcRTL( PacketType ) for _ in range( s.num_terminals )]

    # Components

    s.routers  = [ MeshRouterRTL( PacketType, PositionType )
                   for _ in range( s.num_routers ) ]

    s.channels = [ ChannelRTL( PacketType, latency = chl_lat )
                   for _ in range( num_channels ) ]

    # Wire the position ports of router

    for y in range( nrows ):
      for x in range( ncols ):
        s.routers[y*ncols+x].pos.pos_x //= x
        s.routers[y*ncols+x].pos.pos_y //= y

    # Connect routers together in Mesh
    # NOTE: for now we put all channels in a single list. In the future we
    # may want to divide channels into different groups so that it is
    # easier to configure.

    chl_id  = 0
    for i in range( s.num_routers ):
      if i // ncols > 0:
        s.routers[i].send[SOUTH] //= s.channels[chl_id].recv
        s.channels[chl_id].send  //= s.routers[i-ncols].recv[NORTH]
        chl_id += 1

      if i // ncols < nrows - 1:
        s.routers[i].send[NORTH] //= s.channels[chl_id].recv
        s.channels[chl_id].send  //= s.routers[i+ncols].recv[SOUTH]
        chl_id += 1

      if i % ncols > 0:
        s.routers[i].send[WEST] //= s.channels[chl_id].recv
        s.channels[chl_id].send //= s.routers[i-1].recv[EAST]
        chl_id += 1

      if i % ncols < ncols - 1:
        s.routers[i].send[EAST] //= s.channels[chl_id].recv
        s.channels[chl_id].send //= s.routers[i+1].recv[WEST]
        chl_id += 1

      # Connect the self port (with Network Interface)

      s.recv[i] //= s.routers[i].recv[SELF]
      s.send[i] //= s.routers[i].send[SELF]

      # Connect the unused ports
      # FIXME: for now we hackily ground the payload field so that pymtl
      # won't complain about net need driver.

      if i // ncols == 0:
        s.routers[i].send[SOUTH].rdy         //= 0
        s.routers[i].recv[SOUTH].en          //= 0
        s.routers[i].recv[SOUTH].msg.payload //= 0

      if i // ncols == nrows - 1:
        s.routers[i].send[NORTH].rdy         //= 0
        s.routers[i].recv[NORTH].en          //= 0
        s.routers[i].recv[NORTH].msg.payload //= 0

      if i % ncols == 0:
        s.routers[i].send[WEST].rdy          //= 0
        s.routers[i].recv[WEST].en           //= 0
        s.routers[i].recv[WEST].msg.payload  //= 0

      if i % ncols == ncols - 1:
        s.routers[i].send[EAST].rdy          //= 0
        s.routers[i].recv[EAST].en           //= 0
        s.routers[i].recv[EAST].msg.payload  //= 0

  def line_trace( s ):
    trace    = []
    send_lst = []
    recv_lst = []
    for r in s.routers:
      has_recv = any([ r.recv[i].en for i in range(5) ])
      has_send = any([ r.send[i].en for i in range(5) ])
      if has_send:
        send_lst.append( f'{r.pos}' )
      if has_recv:
        recv_lst.append( f'{r.pos}')
      if has_recv or has_send:
        trace.append( f'  {r.line_trace()}')
    send_str = ','.join( send_lst )
    recv_str = ','.join( recv_lst )
    return f' {send_str:3} -> {recv_str:3}' # + '\n'.join( trace )



#=========================================================================
# TorusNetworkRTL.py
#=========================================================================
# Torus network implementation.
#
# Author : Cheng Tan
#   Date : Mar 10, 2019

from channel.ChannelRTL import ChannelRTL
from ocnlib.ifcs.CreditIfc import (CreditRecvRTL2SendRTL,
                                      RecvRTL2CreditSendRTL)
from pymtl3 import *
from pymtl3.stdlib.ifcs import RecvIfcRTL, SendIfcRTL

from .directions import *
from .TorusRouterRTL import TorusRouterRTL


class TorusNetworkRTL( Component ):

  def construct( s, PacketType, PositionType,
                 ncols=4, nrows=4, chl_lat=0, vc=2, credit_line=2 ):

    # Constants

    s.ncols = ncols
    s.nrows = nrows
    s.num_routers   = ncols * nrows
    s.num_terminals = s.num_routers
    XType           = mk_bits( clog2(ncols) )
    YType           = mk_bits( clog2(nrows ) )

    # Interface

    s.recv = [ RecvIfcRTL(PacketType) for _ in range(s.num_terminals)]
    s.send = [ SendIfcRTL(PacketType) for _ in range(s.num_terminals)]

    # Components

    s.routers = [ TorusRouterRTL( PacketType, PositionType,
                                  ncols=ncols, nrows=nrows,
                                  vc=vc, credit_line=credit_line )
                     for i in range( s.num_routers ) ]

    s.recv_adapters = [ RecvRTL2CreditSendRTL( PacketType, vc=vc,
        credit_line=credit_line ) for _ in range( s.num_routers ) ]
    s.send_adapters = [ CreditRecvRTL2SendRTL( PacketType, vc=vc,
        credit_line=credit_line ) for _ in range( s.num_routers ) ]

    # Connect s.routers together in Torus

    chl_id  = 0
    for i in range (s.num_routers):
      s_idx = (i-nrows+s.num_routers) % s.num_routers
      s.routers[i].send[SOUTH] //= s.routers[s_idx].recv[NORTH]
      s.routers[i].send[NORTH] //=\
        s.routers[(i+ncols+s.num_routers)%s.num_routers].recv[SOUTH]
      s.routers[i].send[WEST]  //=\
        s.routers[i-(i%ncols-(i-1)%ncols)].recv[EAST]
      s.routers[i].send[EAST]  //=\
        s.routers[i+(i+1)%ncols-i%ncols].recv[WEST]

      # Connect the self port (with Network Interface)
      s.recv[i]               //= s.recv_adapters[i].recv
      s.recv_adapters[i].send //= s.routers[i].recv[SELF]

      s.routers[i].send[SELF] //= s.send_adapters[i].recv
      s.send_adapters[i].send //= s.send[i]

#    @s.update
#    def up_pos():
    for y in range( nrows ):
      for x in range( ncols ):
#          idx = y * ncols + x
#          s.routers[idx].pos = PositionType( x, y )
        s.routers[y*ncols+x].pos.pos_x //= XType(x)
        s.routers[y*ncols+x].pos.pos_y //= YType(y)

  def line_trace( s ):
      send_lst = []
      recv_lst = []
      for r in s.routers:
        has_recv = any([ r.recv[i].en for i in range(5) ])
        has_send = any([ r.send[i].en for i in range(5) ])
        if has_send:
          send_lst.append( f'{r.pos}' )
        if has_recv:
          recv_lst.append( f'{r.pos}')
      send_str = ','.join( send_lst )
      recv_str = ','.join( recv_lst )
      return f' {send_str:3} -> {recv_str:3}'


  def elaborate_physical( s ):
    # Initialize dimension for sub-modules.
    BOUNDARY = 10

    for i, r in enumerate( s.routers ):
      r.dim.x = BOUNDARY + i % s.ncols * ( r.dim.w + s.channels[0].dim.w )
      r.dim.y = BOUNDARY + i / s.ncols * ( r.dim.h + s.channels[0].dim.w )

    s.dim.w = 2 * BOUNDARY + s.ncols * ( r.dim.w + s.channels[0].dim.w )
    s.dim.h = 2 * BOUNDARY + s.nrows  * ( r.dim.h + s.channels[0].dim.w )



"""
=========================================================================
BfRouteRTL.py
=========================================================================
RTL router for butterfly network.

Author : Cheng Tan, Yanghui Ou
  Date : April 6, 2019
"""
from pymtl3 import *
from pymtl3.stdlib.ifcs import RecvIfcRTL, SendIfcRTL
from ocnlib.ifcs.PhysicalDimension import PhysicalDimension
from router.InputUnitRTL import InputUnitRTL
from router.OutputUnitRTL import OutputUnitRTL
from router.Router import Router
from router.SwitchUnitRTL import SwitchUnitRTL

from .DTRBflyRouteUnitRTL import DTRBflyRouteUnitRTL


class BflyRouterRTL( Component ):

  # def construct( s,
  #   PacketType,
  #   PositionType,
  #   k_ary = 2,
  #   InputUnitType = InputUnitRTL,
  #   RouteUnitType = DTRBflyRouteUnitRTL,
  #   SwitchUnitType = SwitchUnitRTL
  # ):

  #   super().construct(
  #     PacketType,
  #     PositionType,
  #     k_ary,
  #     k_ary,
  #     InputUnitType,
  #     RouteUnitType,
  #     SwitchUnitType,
  #     OutputUnitRTL,
  #   )

  def construct( s, 
    PacketType, PositionType, k_ary, n_fly,
    InputUnitType  = InputUnitRTL, 
    RouteUnitType  = DTRBflyRouteUnitRTL, 
    SwitchUnitType = SwitchUnitRTL,
    OutputUnitType = OutputUnitRTL,
  ):

    s.dim = PhysicalDimension()
    s.num_inports  = k_ary
    s.num_outports = k_ary

    # Interface

    s.pos  = InPort( PositionType )
    s.recv = [ RecvIfcRTL( PacketType ) for _ in range( s.num_inports  ) ]
    s.send = [ SendIfcRTL( PacketType ) for _ in range( s.num_outports ) ]

    # Components

    s.input_units  = [ InputUnitType( PacketType )
                      for _ in range( s.num_inports ) ]

    s.route_units  = [ RouteUnitType( PacketType, PositionType, s.num_outports, n_fly=n_fly )
                      for i in range( s.num_inports ) ]

    s.switch_units = [ SwitchUnitType( PacketType, s.num_inports )
                      for _ in range( s.num_outports ) ]

    s.output_units = [ OutputUnitType( PacketType )
                      for _ in range( s.num_outports ) ]

    # Connection

    for i in range( s.num_inports ):
      s.recv[i]             //= s.input_units[i].recv
      s.input_units[i].give //= s.route_units[i].get
      s.pos                 //= s.route_units[i].pos

    for i in range( s.num_inports ):
      for j in range( s.num_outports ):
        s.route_units[i].give[j] //= s.switch_units[j].get[i]

    for j in range( s.num_outports ):
      s.switch_units[j].give //= s.output_units[j].get
      s.output_units[j].send //= s.send[j]

  # Line trace

  def line_trace( s ):
    return "{}({}){}".format(
      "|".join( [ f"{x}" for x in s.recv ] ),
      s.pos,
      "|".join( [ f"{x}" for x in s.send ] )
    )

  def elaborate_physical( s ):
    s.dim.w = 50
    s.dim.h = 150



"""
==========================================================================
DORYTorusRouteUnitRTL.py
==========================================================================
A DOR route unit with get/give interface for Torus topology.

Author : Yanghui Ou, Cheng Tan
  Date : June 29, 2019
"""
from copy import deepcopy

from pymtl3 import *
from pymtl3.stdlib.ifcs import GetIfcRTL, GiveIfcRTL

from .directions import *


class DORYTorusRouteUnitRTL( Component ):

  def construct( s, PacketType, PositionType, ncols=2, nrows=2 ):

    # Constants

    s.num_outports = 5
    s.ncols = ncols
    s.nrows = nrows

    # Here we add 1 to avoid overflow

    posx_type     = mk_bits( clog2( ncols ) )
    posy_type     = mk_bits( clog2( nrows ) )
    ns_dist_type  = mk_bits( clog2( nrows+1 ) )
    we_dist_type  = mk_bits( clog2( ncols+1 ) )

    s.last_row_id = ns_dist_type( nrows-1 )
    s.last_col_id = we_dist_type( ncols-1 )

    # Interface

    s.get  = GetIfcRTL( PacketType )
    s.give = [ GiveIfcRTL (PacketType) for _ in range ( s.num_outports ) ]
    s.pos  = InPort( PositionType )

    # Componets

    s.out_dir       = Wire( Bits3        )
    s.give_ens      = Wire( Bits5        )
    s.turning       = Wire( Bits1        )
    s.north_dist    = Wire( ns_dist_type )
    s.south_dist    = Wire( ns_dist_type )
    s.west_dist     = Wire( we_dist_type )
    s.east_dist     = Wire( we_dist_type )
    s.give_msg_wire = Wire( PacketType   )

    # Connections

    for i in range( s.num_outports ):
      s.give_ens[i]   //= s.give[i].en
      s.give_msg_wire //= s.give[i].ret

    # Calculate distance

    @s.update
    def up_ns_dist():
      if s.get.ret.dst_y < s.pos.pos_y:
        s.south_dist = ns_dist_type(s.pos.pos_y) - ns_dist_type(s.get.ret.dst_y)
        s.north_dist = s.last_row_id - ns_dist_type(s.pos.pos_y) + ns_dist_type(1) + ns_dist_type(s.get.ret.dst_y)
      else:
        s.south_dist = ns_dist_type(s.pos.pos_y) + ns_dist_type(1) + s.last_row_id - ns_dist_type(s.get.ret.dst_y)
        s.north_dist = ns_dist_type(s.get.ret.dst_y) - ns_dist_type(s.pos.pos_y)

    @s.update
    def up_we_dist():
      if s.get.ret.dst_x < s.pos.pos_x:
        s.west_dist = we_dist_type(s.pos.pos_x) - we_dist_type(s.get.ret.dst_x)
        s.east_dist = s.last_col_id - we_dist_type(s.pos.pos_x) + we_dist_type(1) + we_dist_type(s.get.ret.dst_x)
      else:
        s.west_dist = we_dist_type(s.pos.pos_x) + we_dist_type(1) + s.last_col_id - we_dist_type(s.get.ret.dst_x)
        s.east_dist = we_dist_type(s.get.ret.dst_x) - we_dist_type(s.pos.pos_x)

    # Routing logic

    @s.update
    def up_ru_routing():

      s.give_msg_wire = deepcopy( s.get.ret )
      s.out_dir = b3(0)
      s.turning = b1(0)

      for i in range( s.num_outports ):
        s.give[i].rdy = b1(0)

      if s.get.rdy:
        if s.pos.pos_x == s.get.ret.dst_x and s.pos.pos_y == s.get.ret.dst_y:
          s.out_dir = SELF
        elif s.get.ret.dst_y != s.pos.pos_y:
          s.out_dir = NORTH if s.north_dist < s.south_dist else SOUTH
        else:
          s.out_dir = WEST if s.west_dist < s.east_dist else EAST

        # Turning logic

        s.turning = ( s.get.ret.src_x == s.pos.pos_x ) & \
                    ( s.get.ret.src_y != s.pos.pos_y ) & \
                    ( s.out_dir == WEST ) | ( s.out_dir == EAST )

        # Dateline logic

        if s.turning:
          s.give_msg_wire.vc_id = b1(0)

        if s.pos.pos_x == posx_type(0) and s.out_dir == WEST:
          s.give_msg_wire.vc_id = b1(1)
        elif s.pos.pos_x == posx_type(s.last_col_id) and s.out_dir == EAST:
          s.give_msg_wire.vc_id = b1(1)
        elif s.pos.pos_y == posy_type(0) and s.out_dir == SOUTH:
          s.give_msg_wire.vc_id = b1(1)
        elif s.pos.pos_y == posy_type(s.last_col_id) and s.out_dir == NORTH:
          s.give_msg_wire.vc_id = b1(1)

        s.give[ s.out_dir ].rdy = b1(1)

    @s.update
    def up_ru_get_en():
      s.get.en = s.give_ens > b5(0)

  # Line trace

  def line_trace( s ):
    out_str = "|".join([ str(s.give[i]) for i in range( s.num_outports ) ])
    dir_str = (
      "N" if s.out_dir == NORTH else
      "S" if s.out_dir == SOUTH else
      "W" if s.out_dir == WEST  else
      "E"
    )
    turn_str = "t" if s.turning else " "
    return f"{s.get}({dir_str},{turn_str}){out_str}"



#=========================================================================
# TorusNetworkRTL_test.py
#=========================================================================
# Test for TorusNetworkRTL
#
# Author : Yanghui Ou, Cheng Tan
#   Date : July 1, 2019

import hypothesis
from hypothesis import strategies as st

from ocnlib.ifcs.packets import mk_mesh_pkt
from ocnlib.ifcs.positions import mk_mesh_pos
from ocnlib.utils import run_sim
from ocnlib.test.net_sinks import TestNetSinkRTL
from pymtl3 import *
from pymtl3.stdlib.test.test_srcs import TestSrcRTL
from torusnet.TorusNetworkFL import torusnet_fl
from torusnet.TorusNetworkRTL import TorusNetworkRTL

#-------------------------------------------------------------------------
# TestHarness
#-------------------------------------------------------------------------

class TestHarness( Component ):

  def construct( s, PktType, ncols, nrows, src_msgs, sink_msgs ):

    s.nrouters = ncols * nrows
    MeshPos = mk_mesh_pos( ncols, nrows )
    match_func = lambda a, b : a.payload == b.payload

    s.srcs  = [ TestSrcRTL   ( PktType, src_msgs[i] )
                for i in range ( s.nrouters ) ]
    s.dut   = TorusNetworkRTL( PktType, MeshPos, ncols, nrows, 0)
    s.sinks = [ TestNetSinkRTL  ( PktType, sink_msgs[i], match_func=match_func )
                for i in range ( s.nrouters ) ]

    # Connections
    for i in range ( s.nrouters ):
      s.srcs[i].send //= s.dut.recv[i]
      s.dut.send[i]  //= s.sinks[i].recv

  def done( s ):
    srcs_done  = True
    sinks_done = True
    for i in range( s.nrouters ):
      if not s.srcs[i].done():
        srcs_done = False
        break
      if not s.sinks[i].done():
        sinks_done = False
        break
    return srcs_done and sinks_done

  def line_trace( s ):
    return s.dut.line_trace()

#-------------------------------------------------------------------------
# mk_src_pkts
#-------------------------------------------------------------------------

def mk_src_pkts( ncols, nrows, lst ):
  nterminals = nrows * ncols
  src_pkts = [ [] for _ in range( nterminals ) ]
  for pkt in lst:
    src_id = int(pkt.src_x) + int(pkt.src_y) * ncols
    src_pkts[ src_id ].append( pkt )
  return src_pkts

#-------------------------------------------------------------------------
# mk_src_pkts
#-------------------------------------------------------------------------

@st.composite
def torus_pkt_strat( draw, ncols, nrows ):
  src_x = draw( st.integers(0, ncols-1), label="src_x" )
  src_y = draw( st.integers(0, nrows-1), label="src_y" )
  dst_x = draw( st.integers(0, ncols-1), label="dst_x" )
  dst_y = draw( st.integers(0, nrows-1), label="dst_y" )
  payload = draw( st.sampled_from([ 0xdeadface, 0xfaceb00c, 0xdeadbabe ]) )
  Pkt = mk_mesh_pkt( ncols, nrows, vc=2 )
  return Pkt( src_x, src_y, dst_x, dst_y, 0, 0, payload )

#=========================================================================
# Test cases
#=========================================================================

class TorusNetwork_Tests:

  def test_simple( s ):
    ncols = 2
    nrows = 2

    Pkt = mk_mesh_pkt( ncols, nrows, vc=2 )

    src_pkts = mk_src_pkts( ncols, nrows, [
      #    src_x  y  dst_x  y   opq vc payload
      Pkt(     1, 0,     0, 1,  0,  0, 0xfaceb00c ),
      Pkt(     1, 1,     1, 0,  0,  0, 0xdeadface ),
    ])
    dst_pkts = torusnet_fl( ncols, nrows, src_pkts )
    th = TestHarness( Pkt, ncols, nrows, src_pkts, dst_pkts )
    run_sim( th )

  def test_simple_3x3( s ):
    ncols = 3
    nrows = 3

    Pkt = mk_mesh_pkt( ncols, nrows, vc=2 )

    src_pkts = mk_src_pkts( ncols, nrows, [
      #    src_x  y  dst_x  y   opq vc payload
      Pkt(     1, 0,     0, 2,  0,  0, 0xfaceb00c ),
      #Pkt(     1, 1,     1, 0,  0,  0, 0xdeadface ),
    ])
    dst_pkts = torusnet_fl( ncols, nrows, src_pkts )
    th = TestHarness( Pkt, ncols, nrows, src_pkts, dst_pkts )
    run_sim( th )

  def test_simple_5x5( s ):
    ncols = 5
    nrows = 5

    Pkt = mk_mesh_pkt( ncols, nrows, vc=2 )

    src_pkts = mk_src_pkts( ncols, nrows, [
      #    src_x  y  dst_x  y   opq vc payload
      Pkt(     1, 0,     0, 4,  0,  0, 0xfaceb00c ),
      #Pkt(     1, 1,     1, 0,  0,  0, 0xdeadface ),
    ])
    dst_pkts = torusnet_fl( ncols, nrows, src_pkts )
    th = TestHarness( Pkt, ncols, nrows, src_pkts, dst_pkts )
    run_sim( th )

  @hypothesis.settings( deadline=None, max_examples=5 )
  # @hypothesis.reproduce_failure('4.24.4', 'AAMDAQEAAAQAAA==') #(1:0)>(0:4)
  @hypothesis.given(
    ncols = st.integers(2, 8),
    nrows = st.integers(2, 8),
    pkts  = st.data(),
  )
  def test_hypothesis( s, ncols, nrows, pkts ):
    Pkt = mk_mesh_pkt( ncols, nrows, vc=2 )

    pkts_lst = pkts.draw(
      st.lists( torus_pkt_strat( ncols, nrows ), max_size=4 ),
      label= "pkts"
    )

    src_pkts = mk_src_pkts( ncols, nrows, pkts_lst )
    dst_pkts = torusnet_fl( ncols, nrows, src_pkts )
    th = TestHarness( Pkt, ncols, nrows, src_pkts, dst_pkts )
    run_sim( th, max_cycles=5000 )



"""
=========================================================================
DORYCMeshRouteUnitRTL.py
=========================================================================
A DOR-Y route unit with get/give interface for CMesh.

Author : Yanghui Ou, Cheng Tan
  Date : Mar 25, 2019
"""
from pymtl3 import *
from pymtl3.stdlib.ifcs import GetIfcRTL, GiveIfcRTL

from .directions import *


class DORYCMeshRouteUnitRTL( Component ):

  def construct( s, PacketType, PositionType, num_outports = 5 ):

    # Constants

    s.num_outports = num_outports
    TType = mk_bits( num_outports )

    # Interface

    s.get  = GetIfcRTL( PacketType )
    s.give = [ GiveIfcRTL (PacketType) for _ in range ( s.num_outports ) ]
    s.pos  = InPort( PositionType )

    # Componets

    s.give_ens = Wire( mk_bits( s.num_outports ) )
    s.give_rdy = [ Wire( Bits1 ) for _ in range( s.num_outports )]

    # Connections

    for i in range( s.num_outports ):
      s.get.ret     //= s.give[i].ret
      s.give_ens[i] //= s.give[i].en
      s.give_rdy[i] //= s.give[i].rdy

    # Routing logic

    @s.update
    def up_ru_routing():

      for i in range( s.num_outports ):
        s.give_rdy[i] = Bits1(0)

      if s.get.rdy:
        if s.pos.pos_x == s.get.ret.dst_x and s.pos.pos_y == s.get.ret.dst_y:
          s.give_rdy[TType(4)+TType(s.get.ret.dst_ter)] = Bits1(1)
        elif s.get.ret.dst_y < s.pos.pos_y:
          s.give_rdy[1] = Bits1(1)
        elif s.get.ret.dst_y > s.pos.pos_y:
          s.give_rdy[0] = Bits1(1)
        elif s.get.ret.dst_x < s.pos.pos_x:
          s.give_rdy[2] = Bits1(1)
        else:
          s.give_rdy[3] = Bits1(1)

    @s.update
    def up_ru_get_en():
      s.get.en = s.give_ens > TType(0)

  # Line trace

  def line_trace( s ):
    out_str = "|".join([ f"{s.give[i]}" for i in range( s.num_outports ) ])
    return f"{s.get}(){out_str}"



"""
==========================================================================
TorusNetworkFL
==========================================================================
Functional level implementation of a torus network.

Author : Yanghui Ou
  Date : July 1, 2019
"""

from pymtl3 import *

#-------------------------------------------------------------------------
# torusnet_fl
#-------------------------------------------------------------------------
# [src_pkts] is a list of lists of packets, where each list of packets
# represents traffic from a certain terminal. The FL model returns
# [dst_pkts], which is also a list of lists of packets, where each list of
# packets represents the output packets of each terminal.

def torusnet_fl( ncols, nrows, src_pkts ):
  nterminals = ncols * nrows
  assert len( src_pkts ) == nterminals

  dst_pkts = [ [] for _ in range( nterminals ) ]

  for packets in src_pkts:
    for pkt in packets:
      # TODO: change vc_id?
      dst_id = int(pkt.dst_x) + int(pkt.dst_y) * ncols
      dst_pkts[ dst_id ].append( pkt )
  return dst_pkts



"""
==========================================================================
sim_utils.py
==========================================================================
Utility functions for pyocn script.

Author: Yanghui Ou
  Date: Sep 24, 2019

"""
import argparse
import os
import sys
import time
from collections import deque
from copy import deepcopy
from dataclasses import dataclass
from random import randint, seed

# Hacky way to add pymtl3-net to path
# TODO: remove this line and use globally installed pymtl3-net
sys.path.insert(0, os.path.dirname( os.path.dirname( os.path.abspath(__file__) ) ) )

from bflynet import BflyNetworkRTL
from cmeshnet import CMeshNetworkRTL
from meshnet import MeshNetworkCL, MeshNetworkRTL
from ocnlib.ifcs.packets import (mk_bfly_pkt, mk_cmesh_pkt, mk_mesh_pkt,
                                    mk_ring_pkt)
from ocnlib.ifcs.positions import mk_bfly_pos, mk_mesh_pos, mk_ring_pos
from ocnlib.sim.CLNetWrapper import CLNetWrapper
from pymtl3 import *
from pymtl3.passes.backends.verilog import TranslationPass, TranslationConfigs
from pymtl3.passes import TracingConfigs
from ringnet import RingNetworkRTL
from torusnet import TorusNetworkRTL

seed( 0xfaceb00c )


#-------------------------------------------------------------------------
# module level variable
#-------------------------------------------------------------------------

verbose = False

#-------------------------------------------------------------------------
# convenience variable
#-------------------------------------------------------------------------

_green = '\33[32m'
_reset = '\033[0m'
_red   = '\033[31m'

#-------------------------------------------------------------------------
# Verbose print
#-------------------------------------------------------------------------

def vprint( msg='', value=None ):
  if verbose:
    if value is not None:
      print( msg, value )
    else:
      print( msg )

#-------------------------------------------------------------------------
# add_mesh_arg
#-------------------------------------------------------------------------

def _add_mesh_arg( p ):
  p.add_argument( '--ncols', type=int, default=2, metavar='', help='Number of columns.' )
  p.add_argument( '--nrows', type=int, default=2, metavar='', help='Number of rows.' )
  p.add_argument( '--channel-lat', type=int, default=0,  metavar='', help='Channel latency in cycles. 0 means combinational channel.' )
  p.add_argument( '--channel-bw',  type=int, default=32, metavar='', help='Channel bandwidth in bits.' )

#-------------------------------------------------------------------------
# add_ring_arg
#-------------------------------------------------------------------------

def _add_ring_arg( p ):
  p.add_argument( '--nterminals',  type=int, default=4, metavar='', help='Number of terminals' )
  p.add_argument( '--channel-lat', type=int, default=0,  metavar='', help='Channel latency. 0 means combinational channel.' )
  p.add_argument( '--channel-bw',  type=int, default=32, metavar='', help='Channel bandwidth in bits.' )

#-------------------------------------------------------------------------
# add_torus_arg
#-------------------------------------------------------------------------

def _add_torus_arg( p ):
  p.add_argument( '--ncols', type=int, default=2, metavar='', help='Number of columns.' )
  p.add_argument( '--nrows', type=int, default=2, metavar='', help='Number of rows.' )
  p.add_argument( '--channel-lat', type=int, default=0,  metavar='', help='Channel latency. 0 means combinational channel.' )
  p.add_argument( '--channel-bw',  type=int, default=32, metavar='', help='Channel bandwidth in bits.' )


#-------------------------------------------------------------------------
# add_cmesh_arg
#-------------------------------------------------------------------------

def _add_cmesh_arg( p ):
  p.add_argument( '--ncols', type=int, default=2, metavar='', help='Number of columns.' )
  p.add_argument( '--nrows', type=int, default=2, metavar='', help='Number of rows.' )
  p.add_argument( '--nterminals-each', type=int, default=2, metavar='', help='Number of terminals per router.' )
  p.add_argument( '--channel-lat', type=int, default=0,  metavar='', help='Channel latency. 0 means combinational channel.' )
  p.add_argument( '--channel-bw',  type=int, default=32, metavar='', help='Channel bandwidth in bits.' )

#-------------------------------------------------------------------------
# add_cmesh_arg
#-------------------------------------------------------------------------

def _add_bfly_arg( p ):
  p.add_argument( '--kary', type=int, default=2, metavar='', help='Radix for each router.' )
  p.add_argument( '--nfly', type=int, default=2, metavar='', help='Number of stages.' )
  p.add_argument( '--channel-lat', type=int, default=0,  metavar='', help='Channel latency. 0 means combinational channel.' )
  p.add_argument( '--channel-bw',  type=int, default=32, metavar='', help='Channel bandwidth in bits.' )


#-------------------------------------------------------------------------
# _mk_mesh_net
#-------------------------------------------------------------------------

def _mk_mesh_net( opts ):
  ncols  = opts.ncols
  nrows  = opts.nrows
  nports = opts.ncols * opts.nrows
  payload_nbits = opts.channel_bw
  channel_lat   = opts.channel_lat

  Pos = mk_mesh_pos( ncols, nrows )
  Pkt = mk_mesh_pkt( ncols, nrows, vc=1, payload_nbits=payload_nbits )
  if hasattr( opts, 'cl' ) and opts.cl:
    cl_net = MeshNetworkCL( Pkt, Pos, ncols, nrows, channel_lat )
    net    = CLNetWrapper( Pkt, cl_net, nports )
  else:
    net = MeshNetworkRTL( Pkt, Pos, ncols, nrows, channel_lat )
  return net

#-------------------------------------------------------------------------
# _mk_ring_net
#-------------------------------------------------------------------------

def _mk_ring_net( opts ):
  nterminals = opts.nterminals
  payload_nbits = opts.channel_bw
  channel_lat   = opts.channel_lat

  Pos = mk_ring_pos( nterminals )
  Pkt = mk_ring_pkt( nterminals, vc=2, payload_nbits=payload_nbits )
  net = RingNetworkRTL( Pkt, Pos, nterminals, channel_lat, vc=2, credit_line=2 )
  return net

#-------------------------------------------------------------------------
# _mk_torus_net
#-------------------------------------------------------------------------

def _mk_torus_net( opts ):
  ncols = opts.ncols
  nrows = opts.nrows
  payload_nbits = opts.channel_bw
  channel_lat   = opts.channel_lat

  Pos = mk_mesh_pos( ncols, nrows )
  Pkt = mk_mesh_pkt( ncols, nrows, vc=2, payload_nbits=payload_nbits )
  net = TorusNetworkRTL( Pkt, Pos, ncols, nrows, channel_lat, vc=2, credit_line=2 )
  return net

#-------------------------------------------------------------------------
# _mk_cmesh_net
#-------------------------------------------------------------------------

def _mk_cmesh_net( opts ):
  ncols = opts.ncols
  nrows = opts.nrows
  payload_nbits = opts.channel_bw
  channel_lat   = opts.channel_lat
  router_ninports  = opts.nterminals_each + 4
  router_noutports = opts.nterminals_each + 4

  Pos = mk_mesh_pos( ncols, nrows )
  Pkt = mk_cmesh_pkt( ncols, nrows, router_ninports, router_noutports,
                      vc=1, payload_nbits=payload_nbits )
  net = CMeshNetworkRTL( Pkt, Pos, ncols, nrows, opts.nterminals_each, channel_lat )
  return net

#-------------------------------------------------------------------------
# _mk_bfly_net
#-------------------------------------------------------------------------

def _mk_bfly_net( opts ):
  kary = opts.kary
  nfly = opts.nfly
  payload_nbits = opts.channel_bw
  channel_lat   = opts.channel_lat

  Pos = mk_bfly_pos( kary, nfly )
  Pkt = mk_bfly_pkt( kary, nfly, vc=1, payload_nbits=payload_nbits )
  net = BflyNetworkRTL( Pkt, Pos, kary, nfly, channel_lat )
  # net.set_param( "top.routers*.construct", k_ary=kary )
  # net.set_param( "top.routers*.route_units*.construct", n_fly=nfly )

  return net

#-------------------------------------------------------------------------
# _gen_dst_id
#-------------------------------------------------------------------------

def _gen_dst_id( pattern, nports, src_id ):
  if pattern == 'urandom':
    return randint( 0, nports-1 )
  elif pattern == 'partition':
    return randint( 0, nports-1 ) & ( nports//2 - 1 ) | ( src_id & ( nports//2 ) )
  elif pattern == 'opposite':
    return ( src_id + nports//2 ) % nports
  elif pattern == 'neighbor':
    return ( src_id + 1 ) % nports
  elif pattern == 'complement':
    return ( nports-1 ) - src_id
  else:
    raise Exception( f'Unkonwn traffic pattern {pattern}' )

#-------------------------------------------------------------------------
# _gen_mesh_net
#-------------------------------------------------------------------------

def _gen_mesh_pkt( opts, timestamp, src_id ):
  ncols = opts.ncols
  nrows = opts.nrows
  payload_nbits = opts.channel_bw
  nports = ncols * nrows

  x_type = mk_bits( clog2( opts.ncols ) )
  y_type = mk_bits( clog2( opts.nrows ) )

  pkt = mk_mesh_pkt( ncols, nrows, vc=1, payload_nbits=payload_nbits )()
  pkt.payload = timestamp

  dst_id    = _gen_dst_id( opts.pattern, nports, src_id )
  pkt.src_x = x_type( src_id %  ncols )
  pkt.src_y = y_type( src_id // ncols )
  pkt.dst_x = x_type( dst_id %  ncols )
  pkt.dst_y = y_type( dst_id // ncols )

  return pkt

#-------------------------------------------------------------------------
# _gen_ring_net
#-------------------------------------------------------------------------

def _gen_ring_pkt( opts, timestamp, src_id ):
  payload_nbits = opts.channel_bw
  nports = opts.nterminals

  id_type = mk_bits( clog2( nports ) )

  pkt = mk_ring_pkt( nports, vc=2, payload_nbits=payload_nbits )()
  pkt.payload = timestamp

  dst_id  = _gen_dst_id( opts.pattern, nports, src_id )
  pkt.src = id_type( src_id )
  pkt.dst = id_type( dst_id )

  return pkt

#-------------------------------------------------------------------------
# _gen_torus_net
#-------------------------------------------------------------------------

def _gen_torus_pkt( opts, timestamp, src_id ):
  ncols = opts.ncols
  nrows = opts.nrows
  payload_nbits = opts.channel_bw
  nports = ncols * nrows

  x_type = mk_bits( clog2( opts.ncols ) )
  y_type = mk_bits( clog2( opts.nrows ) )

  pkt = mk_mesh_pkt( ncols, nrows, vc=2, payload_nbits=payload_nbits )()
  pkt.payload = timestamp

  dst_id    = _gen_dst_id( opts.pattern, nports, src_id )
  pkt.src_x = x_type( src_id %  ncols )
  pkt.src_y = y_type( src_id // ncols )
  pkt.dst_x = x_type( dst_id %  ncols )
  pkt.dst_y = y_type( dst_id // ncols )
  pkt.vc_id = b1(0)

  return pkt

#-------------------------------------------------------------------------
# _gen_cmesh_net
#-------------------------------------------------------------------------

def _gen_cmesh_pkt( opts, timestamp, src_id ):
  ncols = opts.ncols
  nrows = opts.nrows
  nterminals_each = opts.nterminals_each
  payload_nbits   = opts.channel_bw
  nports          = ncols * nrows * nterminals_each

  router_ninports  = opts.nterminals_each + 4
  router_noutports = opts.nterminals_each + 4

  x_type = mk_bits( clog2( ncols ) )
  y_type = mk_bits( clog2( nrows ) )
  t_type = mk_bits( clog2( nterminals_each ) )

  pkt = mk_cmesh_pkt( ncols, nrows, router_ninports, router_noutports,
                      vc=1, payload_nbits=payload_nbits )()
  pkt.payload = timestamp

  dst_id      = _gen_dst_id( opts.pattern, nports, src_id )
  pkt.src_x   = x_type( ( src_id//nterminals_each ) %  ncols )
  pkt.src_y   = y_type( ( src_id//nterminals_each ) // ncols )
  pkt.dst_x   = x_type( ( dst_id//nterminals_each ) %  ncols )
  pkt.dst_y   = y_type( ( dst_id//nterminals_each ) // ncols )
  pkt.dst_ter = t_type( dst_id % nterminals_each )

  return pkt

#-------------------------------------------------------------------------
# _gen_bfly_net
#-------------------------------------------------------------------------

def _gen_bfly_pkt( opts, timestamp, src_id ):
  kary          = opts.kary
  nfly          = opts.nfly
  payload_nbits = opts.channel_bw
  nports        = opts.kary ** opts.nfly

  id_type = mk_bits( clog2( nports ) )

  pkt = mk_bfly_pkt( kary, nfly, vc=1, payload_nbits=payload_nbits )()
  pkt.payload = timestamp

  dst_id  = _gen_dst_id( opts.pattern, nports, src_id )
  pkt.src = id_type( src_id )
  pkt.dst = id_type( dst_id )

  return pkt

#-------------------------------------------------------------------------
# dictionaries
#-------------------------------------------------------------------------

_net_arg_dict = {
  'mesh'  : _add_mesh_arg,
  'ring'  : _add_ring_arg,
  'torus' : _add_torus_arg,
  'cmesh' : _add_cmesh_arg,
  'bfly'  : _add_bfly_arg,
}

_net_inst_dict = {
  'mesh'  : _mk_mesh_net,
  'ring'  : _mk_ring_net,
  'torus' : _mk_torus_net,
  'cmesh' : _mk_cmesh_net,
  'bfly'  : _mk_bfly_net,
}

_net_nports_dict = {
  'mesh'  : lambda opts: opts.ncols * opts.nrows,
  'ring'  : lambda opts: opts.nterminals,
  'torus' : lambda opts: opts.ncols * opts.nrows,
  'cmesh' : lambda opts: opts.ncols * opts.nrows * opts.nterminals_each,
  'bfly'  : lambda opts: opts.kary ** opts.nfly,
}

_pkt_gen_dict = {
  'mesh'  : _gen_mesh_pkt,
  'ring'  : _gen_ring_pkt,
  'torus' : _gen_torus_pkt,
  'cmesh' : _gen_cmesh_pkt,
  'bfly'  : _gen_bfly_pkt,
}

#-------------------------------------------------------------------------
# mk_net_arg_parser
#-------------------------------------------------------------------------

def mk_net_arg_parser( topo ):
  if not topo in _net_arg_dict:
    raise Exception( f'Unknown network topology {topo}' )

  p = argparse.ArgumentParser(
    # description = f'{topo}',
    usage = f'./pymtl-net sim {topo} [<flags>]',
  )
  _net_arg_dict[ topo ]( p )
  return p

#-------------------------------------------------------------------------
# mk_net_inst
#-------------------------------------------------------------------------

def mk_net_inst( topo, opts ):
  if not topo in _net_arg_dict:
    raise Exception( f'Unkonwn network topology {topo}' )
  return _net_inst_dict[ topo ]( opts )

#-------------------------------------------------------------------------
# get_nports
#-------------------------------------------------------------------------

def get_nports( topo,  opts ):
  if not topo in _net_arg_dict:
    raise Exception( f'Unkonwn network topology {topo}' )
  return _net_nports_dict[ topo ]( opts )

#-------------------------------------------------------------------------
# SimResult
#-------------------------------------------------------------------------

@dataclass
class SimResult:
  injection_rate : int   = 0
  avg_latency    : float = 0.0
  pkt_generated  : int   = 0
  mpkt_received  : int   = 0
  total_received : int   = 0
  sim_ncycles    : int   = 0
  elapsed_time   : float = 0.0
  timeout        : bool  = False

  def print_result( self ):
    print( f'injection_rate    : {self.injection_rate} %' )
    print( f'average latency   : {self.avg_latency:.2f}'   )
    print( f'simulated cycles  : {self.sim_ncycles}'       )
    print( f'packets generated : {self.total_generated}'   )
    print( f'packets received  : {self.total_received}'    )
    print( f'#measure packets  : {self.mpkt_received}'     )
    print( f'elapsed time      : {self.elapsed_time:.2f} sec'  )

  def to_row( self ):
    return f'| {self.injection_rate:4} | {self.avg_latency:8.2f} | {self.sim_ncycles/self.elapsed_time:5.1f} |'

#-------------------------------------------------------------------------
# net_simulate
#-------------------------------------------------------------------------
# TODO: dump vcd

def net_simulate( topo, opts ):
  if not topo in _net_arg_dict:
    raise Exception( f'Unkonwn network topology {topo}' )

  # Metadata
  nports = get_nports( topo, opts )
  p_type = mk_bits( opts.channel_bw )

  # Instantiate network instance
  vprint( f' - instantiating {topo} with {nports} terminals')
  net = mk_net_inst( topo, opts )

  # Infinite source queues
  src_q = [ deque() for _ in range( nports ) ]

  if opts.dump_vcd:
    vprint( f' - enabling vcd dumping' )
    net.config_tracing = TracingConfigs( tracing='vcd', vcd_file_name=f'{topo}-{nports}-{opts.injection_rate}' )

  # Elaborating network instance
  vprint( f' - elaborating {topo}' )
  net.elaborate()
  net.apply( SimulationPass() )

  vprint( f' - resetting network')
  net.sim_reset()
  net.tick()

  # Constants

  warmup_ncycles   = opts.warmup_ncycles
  measure_npackets = opts.measure_npackets * 2
  timeout_ncycles  = opts.timeout_ncycles


  vprint( f' - simulation starts' )
  vprint( f' - #measure generated / #measure injected / total generated ---- #measure received / total received')

  injection_rate  = opts.injection_rate
  ncycles         = 0
  total_generated = 0
  mpkt_generated  = 0
  mpkt_received   = 0
  total_received  = 0
  total_latency   = 0
  all_received    = 0
  mpkt_injected   = 0

  for i in range( nports ):
    net.send[i].rdy = b1(1) # Always ready

  # Run simulation

  start_time = time.monotonic()

  while True:
    for i in range( nports ):

      # Inject packets to source queue
      if randint(1,100) <= injection_rate:

        # TODO: use number of packets
        # Warmup phase - inject non-measure packet
        if ncycles <= warmup_ncycles:
          # FIXME: we may want to convert ncycles to bits
          pkt = _pkt_gen_dict[ topo ]( opts, p_type(0), i )

        # Sample phase - inject measure packet
        elif mpkt_generated < measure_npackets:
          pkt = _pkt_gen_dict[ topo ]( opts, b32(ncycles), i )
          mpkt_generated += 1

        # Drain phase - just inject
        else:
          pkt = _pkt_gen_dict[ topo ]( opts, p_type(0), i )

        total_generated += 1
        src_q[i].append( pkt )

      # Inject packets from source queue to network
      if len( src_q[i] ) > 0 and net.recv[i].rdy:
        recv_pkt = src_q[i].popleft()
        net.recv[i].msg = recv_pkt
        net.recv[i].en  = b1(1)
        if int(recv_pkt.payload) > 0:
          mpkt_injected += 1
      else:
        net.recv[i].en  = b1(0)

      # Receive packets from network
      if net.send[i].en:
        total_received += 1
        if int(net.send[i].msg.payload) > 0:
          timestamp = int(net.send[i].msg.payload)
          total_latency += ( ncycles - timestamp )
          mpkt_received += 1

      # Check finish

      if mpkt_received >= opts.measure_npackets and mpkt_generated == measure_npackets:
        elapsed_time = time.monotonic() - start_time
        result = SimResult()
        result.injection_rate  = injection_rate
        result.avg_latency     = float( total_latency ) / mpkt_received
        result.total_generated = total_generated
        result.mpkt_received   = mpkt_received
        result.total_received  = total_received
        result.sim_ncycles     = ncycles
        result.elapsed_time    = elapsed_time
        return result

      elif ncycles >= timeout_ncycles:
        elapsed_time = time.monotonic() - start_time
        result = SimResult()
        result.injection_rate  = injection_rate
        result.avg_latency     = float( total_latency ) / mpkt_received
        result.total_generated = total_generated
        result.mpkt_received   = mpkt_received
        result.total_received  = total_received
        result.sim_ncycles     = ncycles
        result.elapsed_time    = elapsed_time
        result.timeout         = True
        vprint( ' - TIMEOUT!' )
        return result

    # Advance simulation

    if opts.trace:
      print( f'{ncycles:3}: {net.line_trace()}' )

    if opts.verbose and ncycles % 100 == 1:
      print( f'{ncycles:4}: gen {mpkt_generated}/{mpkt_injected}/{total_generated} recv {mpkt_received}/{total_received}' )

    net.tick()
    ncycles += 1

#-------------------------------------------------------------------------
# net_simulate_cl
#-------------------------------------------------------------------------

def net_simulate_cl( topo, opts ):
  if not topo in _net_arg_dict:
    raise Exception( f'Unkonwn network topology {topo}' )

  # Metadata
  nports = get_nports( topo, opts )
  p_type = mk_bits( opts.channel_bw )

  # Instantiate network instance
  vprint( f' - instantiating {topo} with {nports} terminals')
  net = mk_net_inst( topo, opts )

  # Infinite source queues
  src_q = [ deque() for _ in range( nports ) ]

  # Elaborating network instance
  vprint( f' - elaborating {topo}' )
  net.elaborate()
  net.apply( SimulationPass() )

  vprint( f' - resetting network')
  net.sim_reset()
  net.tick()

  # Constants

  warmup_ncycles   = opts.warmup_ncycles
  measure_npackets = opts.measure_npackets * 2
  timeout_ncycles  = opts.timeout_ncycles

  vprint( f' - simulation starts' )
  vprint( f' - #measure generated / #measure injected / total generated  #measure received / total received')
  injection_rate  = max( 1, opts.injection_rate )
  ncycles         = 0
  total_generated = 0
  mpkt_generated  = 0
  mpkt_received   = 0
  total_received  = 0
  total_latency   = 0
  all_received    = 0
  mpkt_injected   = 0

  # Run simulation

  start_time = time.monotonic()

  while True:
    for i in range( nports ):

      # Inject packets to source queue
      if randint(1,100) <= injection_rate:

        # TODO: use number of packets
        # Warmup phase - inject non-measure packet
        if ncycles <= warmup_ncycles:
          # FIXME: we may want to convert ncycles to bits
          pkt = _pkt_gen_dict[ topo ]( opts, p_type(0), i )

        # Sample phase - inject measure packet
        elif mpkt_generated < measure_npackets:
          pkt = _pkt_gen_dict[ topo ]( opts, b32(ncycles), i )
          mpkt_generated += 1

        # Drain phase - just inject
        else:
          pkt = _pkt_gen_dict[ topo ]( opts, p_type(0), i )

        total_generated += 1
        src_q[i].append( pkt )

      # Inject packets from source queue to network
      if len( src_q[i] ) > 0 and net.recv[i].rdy():
        recv_pkt = src_q[i].popleft()
        net.recv[i]( recv_pkt )
        if int(recv_pkt.payload) > 0:
          mpkt_injected += 1

      # Receive packets from network
      if net.give[i].rdy():
        total_received += 1
        recv_pkt = net.give[i]()
        if int(recv_pkt.payload) > 0:
          timestamp = int(recv_pkt.payload)
          total_latency += ( ncycles - timestamp )
          mpkt_received += 1

      # Check finish

      if mpkt_received >= opts.measure_npackets and mpkt_generated == measure_npackets:
        elapsed_time = time.monotonic() - start_time
        result = SimResult()
        result.injection_rate  = injection_rate
        result.avg_latency     = float( total_latency ) / mpkt_received
        result.total_generated = total_generated
        result.mpkt_received   = mpkt_received
        result.total_received  = total_received
        result.sim_ncycles     = ncycles
        result.elapsed_time    = elapsed_time
        return result

      elif ncycles >= timeout_ncycles:
        elapsed_time = time.monotonic() - start_time
        result = SimResult()
        result.injection_rate  = injection_rate
        result.avg_latency     = float( total_latency ) / mpkt_received
        result.total_generated = total_generated
        result.mpkt_received   = mpkt_received
        result.total_received  = total_received
        result.sim_ncycles     = ncycles
        result.elapsed_time    = elapsed_time
        result.timeout         = True
        vprint( ' - TIMEOUT!' )
        return result

    # Advance simulation

    if opts.trace:
      print( f'{ncycles:3}: {net.line_trace()}' )

    if opts.verbose and ncycles % 100 == 1:
      print( f'{ncycles:4}: gen {mpkt_generated}/{mpkt_injected}/{total_generated} recv {mpkt_received}/{total_received}' )

    net.tick()
    ncycles += 1

#-------------------------------------------------------------------------
# net_simulate_sweep
#-------------------------------------------------------------------------

def net_simulate_sweep( topo, opts ):

  print( f'Topology: {topo} Pattern: {opts.pattern}' )

  result_lst = []

  cur_inj       = 0
  pre_inj       = 0
  cur_avg_lat   = 0.0
  pre_avg_lat   = 0.0
  zero_load_lat = 0.0
  slope         = 0.0
  step          = opts.sweep_step
  threshold    = opts.sweep_thresh

  sim_func = net_simulate if not opts.cl else net_simulate_cl

  while cur_avg_lat <= threshold and cur_inj <= 100:
    new_opts = deepcopy( opts )
    new_opts.injection_rate = max( 1, cur_inj )

    vprint()
    vprint( '-'*74 )
    vprint( f'injection_rate : {cur_inj} %' )
    vprint( '-'*74 )
    vprint()

    result = sim_func( topo, new_opts )
    result_lst.append( result )

    if opts.verbose:
      result.print_result()

    cur_avg_lat = result.avg_latency

    if cur_inj == 0:
      zero_load_lat = cur_avg_lat

    else:
      slope = ( cur_avg_lat - pre_avg_lat ) / ( cur_inj - pre_inj )
      if slope >= 1.0:
        step = max( 1, step // 2 )

    pre_inj =  cur_inj
    cur_inj += step
    pre_avg_lat = cur_avg_lat

  # Print table TODO: more informative?

  print()
  print( '+------+----------+-------+' )
  print( '| inj% | avg. lat | speed |' )
  print( '+------+----------+-------+' )

  for r in result_lst:
    print( f'{r.to_row()}' )

  print( '+------+----------+-------+' )

#-------------------------------------------------------------------------
# gen_verilog
#-------------------------------------------------------------------------

def gen_verilog( topo, opts ):
  os.system(f'[ ! -e {topo}.sv ] || rm {topo}.sv')

  vprint( f' - instantiating {topo}')
  net = mk_net_inst( topo, opts )

  vprint( f' - elaborating {topo}')
  net.elaborate()

  vprint( f' - applying translation pass' )
  net.config_verilog_translate = TranslationConfigs( explicit_module_name = f'{topo}' )
  net.apply( TranslationPass() )

  os.system(f'mv {net.translated_top_module_name}.v {topo}.v')

#-------------------------------------------------------------------------
# smoke_test
#-------------------------------------------------------------------------

def smoke_test( topo, opts ):
  if not topo in _net_arg_dict:
    raise Exception( f'Unkonwn network topology {topo}' )

  # Metadata
  nports = get_nports( topo, opts )
  p_type = mk_bits( opts.channel_bw )

  # Instantiate network instance
  vprint( f' - instantiating {topo} with {nports} terminals')
  net = mk_net_inst( topo, opts )

  if opts.dump_vcd:
    vprint( f' - enabling vcd dumping' )
    net.config_tracing = TracingConfigs( tracing='vcd', vcd_file_name=f'{topo}-{nports}-test' )

  # Elaborating network instance
  vprint( f' - elaborating {topo}' )
  net.elaborate()
  net.apply( SimulationPass() )

  vprint( f' - resetting network')
  net.sim_reset()
  net.tick()

  vprint( f' - simulation starts' )
  ncycles = 0

  for i in range( nports ):
    net.send[i].rdy = b1(1) # Always ready

  # Inject a packet to port 0
  while not net.recv[0].rdy:
    if opts.trace:
      print( f'{ncycles:3}: {net.line_trace()}' )
    net.tick()
    ncycles += 1

  pkt_opts = deepcopy( opts )
  pkt_opts.pattern = 'complement'
  pkt = _pkt_gen_dict[ topo ]( pkt_opts, p_type(1024), 0 )
  net.recv[0].msg = pkt
  net.recv[0].en  = b1(1)

  # Tick one cycle and stops injecting
  if opts.trace:
    print( f'{ncycles:3}: {net.line_trace()}' )
  net.tick()
  ncycles += 1
  net.recv[0].en = b1(0)

  # Wait until packets arrives
  while not net.send[ nports-1 ].en:
    if opts.trace:
      print( f'{ncycles:3}: {net.line_trace()}' )
    net.tick()
    ncycles += 1

  if net.send[ nports-1 ].msg.payload == p_type(1024):
    print( f' - [{_green}passed{_reset}]' )
  else:
    erint( f' - [{_red}FAILED{_reset}]' )

  # Extra ticks to improve waveform
  net.tick()
  net.tick()
  net.tick()
